from __future__ import annotations

import io
from binascii import crc32
from typing import BinaryIO

from dissect.util.stream import OverlayStream

HEADER_FOOTER_SIZE = 12
CRC_SIZE = 4


def repair_checksum(fh: BinaryIO) -> BinaryIO:
    """Repair CRC32 checksums for all headers in an XZ stream.

    FortiOS XZ files have (on purpose) corrupt streams which they read using a modified ``xz`` binary.
    The only thing changed are the CRC32 checksums, so partially parse the XZ file and fix all of them.

    References:
        - https://tukaani.org/xz/xz-file-format-1.1.0.txt
        - https://github.com/Rogdham/python-xz

    Args:
        fh: A file-like object of an LZMA stream to repair.
    """
    file_size = fh.seek(0, io.SEEK_END)
    repaired = OverlayStream(fh, file_size)
    fh.seek(0)

    header = fh.read(HEADER_FOOTER_SIZE)
    # Check header magic
    magic = b"\xfd7zXZ\x00"
    if header[: len(magic)] != magic:
        raise ValueError("Not an XZ file")

    # Add correct header CRC32
    repaired.add(fh.tell() - CRC_SIZE, _crc32(header[len(magic) : HEADER_FOOTER_SIZE - CRC_SIZE]))

    footer_offset = fh.seek(-HEADER_FOOTER_SIZE, io.SEEK_END)
    footer = fh.read(HEADER_FOOTER_SIZE)

    # Check footer magic
    footer_magic = b"YZ"
    if footer[HEADER_FOOTER_SIZE - len(footer_magic) : HEADER_FOOTER_SIZE] != footer_magic:
        raise ValueError("Not an XZ file")

    # Add correct footer CRC32
    repaired.add(footer_offset, _crc32(footer[CRC_SIZE : HEADER_FOOTER_SIZE - len(footer_magic)]))

    backward_size = (int.from_bytes(footer[4:8], "little") + 1) * 4
    fh.seek(-HEADER_FOOTER_SIZE - backward_size, io.SEEK_END)
    index = fh.read(backward_size)

    # Add correct index CRC32
    repaired.add(fh.tell() - CRC_SIZE, _crc32(index[:-CRC_SIZE]))

    # Parse the index
    isize, num_records = _mbi(index[1:])
    index = index[1 + isize : -4]
    records = []
    for _ in range(num_records):
        if not index:
            raise ValueError("Missing index size")

        isize, unpadded_size = _mbi(index)
        if not unpadded_size:
            raise ValueError("Missing index record unpadded size")

        index = index[isize:]
        if not index:
            raise ValueError("Missing index size")

        isize, uncompressed_size = _mbi(index)
        if not uncompressed_size:
            raise ValueError("Missing index record uncompressed size")

        index = index[isize:]
        records.append((unpadded_size, uncompressed_size))

    block_start = file_size - HEADER_FOOTER_SIZE - backward_size
    blocks_len = sum((unpadded_size + 3) & ~3 for unpadded_size, _ in records)
    block_start -= blocks_len

    # Iterate over all blocks and add the correct block header CRC32
    for unpadded_size, _ in records:
        fh.seek(block_start)

        block_header = fh.read(1)
        block_header_size = (block_header[0] + 1) * 4
        block_header += fh.read(block_header_size - 1)
        repaired.add(fh.tell() - CRC_SIZE, _crc32(block_header[:-CRC_SIZE]))

        block_start += (unpadded_size + 3) & ~3

    return repaired


def _mbi(data: bytes) -> tuple[int, int]:
    """Decode a multibyte integer.

    The encoding is similar to most other "varint" encodings. For each byte, the 7 least significant bits are used for
    the integer value. The most significant bit is used to indicate if the integer continues in the next byte.
    Bytes are ordered in little endian byte order, meaning the least significant byte comes first.
    """
    value = 0
    for size, byte in enumerate(data):
        value |= (byte & 0x7F) << (size * 7)
        if not byte & 0x80:
            return size + 1, value
    raise ValueError("Invalid mbi")


def _crc32(data: bytes) -> bytes:
    return int.to_bytes(crc32(data), CRC_SIZE, "little")



# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-frs2/8cb5bae9-edf3-4833-9f0a-9d7e24218d3d
# https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-XCA/[MS-XCA].pdf
from __future__ import annotations

import io
import struct
from typing import BinaryIO, NamedTuple


class Symbol(NamedTuple):
    length: int
    symbol: int


def _read_16_bit(fh: BinaryIO) -> int:
    return struct.unpack("<H", fh.read(2).rjust(2, b"\x00"))[0]


class Node:
    __slots__ = ("children", "is_leaf", "symbol")

    def __init__(self, symbol: Symbol | None = None, is_leaf: bool = False):
        self.symbol = symbol
        self.is_leaf = is_leaf
        self.children = [None, None]


def _add_leaf(nodes: list[Node], idx: int, mask: int, bits: int) -> int:
    node = nodes[0]
    i = idx + 1

    while bits > 1:
        bits -= 1
        childidx = (mask >> bits) & 1
        if node.children[childidx] is None:
            node.children[childidx] = nodes[i]
            nodes[i].is_leaf = False
            i += 1
        node = node.children[childidx]

    node.children[mask & 1] = nodes[idx]
    return i


def _build_tree(buf: bytes) -> Node:
    if len(buf) != 256:
        raise ValueError("Not enough data for Huffman code tree")

    nodes = [Node() for _ in range(1024)]
    symbols: list[Symbol] = []

    for i, c in enumerate(buf):
        symbols.append(Symbol(c & 0x0F, i * 2))
        symbols.append(Symbol((c >> 4) & 0x0F, i * 2 + 1))

    symbols = sorted(symbols)

    symbol_index_start = 0
    for s in symbols:
        if s.length > 0:
            break
        symbol_index_start += 1

    mask = 0
    bits = 1

    root = nodes[0]

    tree_index = 1
    for symbol_index in range(symbol_index_start, 512):
        s = symbols[symbol_index]

        node = nodes[tree_index]
        node.symbol = s.symbol
        node.is_leaf = True

        mask = (mask << s.length - bits) & 0xFFFFFFFF
        bits = s.length

        tree_index = _add_leaf(nodes, tree_index, mask, bits)
        mask += 1

    return root


class BitString:
    def __init__(self):
        self.source = None
        self.mask = 0
        self.bits = 0

    @property
    def index(self) -> int:
        return self.source.tell()

    def init(self, fh: BinaryIO) -> None:
        self.mask = (_read_16_bit(fh) << 16) + _read_16_bit(fh)
        self.bits = 32
        self.source = fh

    def read(self, n: int) -> bytes:
        return self.source.read(n)

    def lookup(self, n: int) -> int:
        if n == 0:
            return 0

        return self.mask >> (32 - n)

    def skip(self, n: int) -> None:
        self.mask = (self.mask << n) & 0xFFFFFFFF
        self.bits -= n
        if self.bits < 16:
            self.mask += _read_16_bit(self.source) << (16 - self.bits)
            self.bits += 16

    def decode(self, root: Node) -> Symbol:
        node = root
        while not node.is_leaf:
            bit = self.lookup(1)
            self.skip(1)
            node = node.children[bit]
        return node.symbol


def decompress(src: bytes | BinaryIO) -> bytes:
    """LZXPRESS decompress from a file-like object or bytes.

    Decompresses until EOF of the input data.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    dst = bytearray()

    start_offset = src.tell()
    src.seek(0, io.SEEK_END)
    size = src.tell() - start_offset
    src.seek(start_offset, io.SEEK_SET)

    bitstring = BitString()

    while src.tell() - start_offset < size:
        root = _build_tree(src.read(256))
        bitstring.init(src)

        chunk_size = 0
        while chunk_size < 65536 and src.tell() - start_offset < size:
            symbol = bitstring.decode(root)
            if symbol < 256:
                dst.append(symbol)
                chunk_size += 1
            else:
                symbol -= 256
                length = symbol & 0x0F
                symbol >>= 4

                offset = (1 << symbol) + bitstring.lookup(symbol)

                if length == 15:
                    length = ord(bitstring.read(1)) + 15

                    if length == 270:
                        length = _read_16_bit(bitstring.source)

                bitstring.skip(symbol)

                length += 3

                remaining = length
                while remaining > 0:
                    match_size = min(remaining, offset)
                    dst += dst[-offset : (-offset + match_size) or None]
                    remaining -= match_size

                chunk_size += length

    return bytes(dst)



from __future__ import annotations

import io
import os
import sys
import zlib
from bisect import bisect_left, bisect_right
from threading import Lock
from typing import BinaryIO

STREAM_BUFFER_SIZE = int(os.getenv("DISSECT_STREAM_BUFFER_SIZE", io.DEFAULT_BUFFER_SIZE))


class AlignedStream(io.RawIOBase):
    """Basic buffered stream that provides aligned reads.

    Must be subclassed for various stream implementations. Subclasses can implement:
        - :meth:`~AlignedStream._read`
        - :meth:`~AlignedStream._seek`

    The offset and length for ``_read`` are guaranteed to be aligned for streams of a known size.
    If your stream has an unknown size (i.e. ``size == None``), reads of length ``-1`` (i.e. read until EOF) will be
    passed through to your implementation of ``_read``.
    The only time that overriding ``_seek`` would make sense is if there's no known size of your stream,
    but still want to provide ``SEEK_END`` functionality.

    Most subclasses of ``AlignedStream`` take one or more file-like objects as source.
    Operations on these subclasses, like reading, will modify the source file-like object as a side effect.

    Args:
        size: The size of the stream. This is used in read and seek operations. ``None`` if unknown.
        align: The alignment size. Read operations are aligned on this boundary. Also determines buffer size.

    .. automethod:: _read
    .. automethod:: _seek
    """

    def __init__(self, size: int | None = None, align: int = STREAM_BUFFER_SIZE):
        super().__init__()
        self.size = size
        self.align = align

        self._pos = 0
        self._pos_align = 0

        self._buf = None
        self._lock = Lock()

    def readable(self) -> bool:
        """Indicate that the stream is readable."""
        return True

    def seekable(self) -> bool:
        """Indicate that the stream is seekable."""
        return True

    def seek(self, pos: int, whence: int = io.SEEK_SET) -> int:
        """Seek the stream to the specified position.

        Returns:
            The new stream position after seeking.
        """
        with self._lock:
            pos = self._seek(pos, whence)
            self._set_pos(pos)

            return pos

    def _seek(self, pos: int, whence: int = io.SEEK_SET) -> int:
        """Calculate and return the new stream position after a seek."""
        if whence == io.SEEK_SET:
            if pos < 0:
                raise ValueError(f"negative seek position {pos}")
        elif whence == io.SEEK_CUR:
            pos = max(0, self._pos + pos)
        elif whence == io.SEEK_END:
            if self.size is None:
                raise IOError("invalid whence value for stream with no size")
            pos = max(0, self.size + pos)
        else:
            raise IOError("invalid whence value")

        return pos

    def _set_pos(self, pos: int) -> None:
        """Update the position and aligned position within the stream."""
        new_pos_align = pos - (pos % self.align)

        if self._pos_align != new_pos_align:
            self._pos_align = new_pos_align
            self._buf = None

        self._pos = pos

    def tell(self) -> int:
        """Return current stream position."""
        return self._pos

    def _fill_buf(self) -> None:
        """Fill the alignment buffer if we can."""
        if self._buf or (self.size is not None and (self.size <= self._pos or self.size <= self._pos_align)):
            # Don't fill the buffer if:
            # - We already have a buffer
            # - The stream position is at the end (or beyond) the stream size
            return

        self._buf = self._read(self._pos_align, self.align)

    def read(self, n: int = -1) -> bytes:
        """Read and return up to ``n`` bytes, or read to the end of the stream if ``n`` is ``-1``.

        Returns an empty bytes object on EOF.
        """
        if n is not None and n < -1:
            raise ValueError("invalid number of bytes to read")

        r = []
        size = self.size
        align = self.align

        with self._lock:
            if size is None and n == -1:
                r = []
                if self._buf:
                    buffer_pos = self._pos - self._pos_align
                    r.append(self._buf[buffer_pos:])
                    self._set_pos(self._pos_align + align)

                r.append(self._read(self._pos_align, -1))

                buf = b"".join(r)
                self._set_pos(self._pos + len(buf))
                return buf

            # If we know the stream size, adjust n
            if size is not None:
                remaining = size - self._pos
                n = remaining if n == -1 else min(n, remaining)

            # Short path for when it turns out we don't need to read anything
            if n == 0 or (size is not None and size <= self._pos):
                return b""

            # Read misaligned start from buffer
            if self._pos != self._pos_align:
                self._fill_buf()

                buffer_pos = self._pos - self._pos_align
                remaining = align - buffer_pos
                buffer_len = min(n, remaining)

                r.append(self._buf[buffer_pos : buffer_pos + buffer_len])

                n -= buffer_len
                self._set_pos(self._pos + buffer_len)

            # Aligned blocks
            if n >= align:
                count, n = divmod(n, align)

                read_len = count * align
                r.append(self._read(self._pos, read_len))

                self._set_pos(self._pos + read_len)

            # Misaligned remaining bytes
            if n > 0:
                self._fill_buf()
                r.append(self._buf[:n])
                self._set_pos(self._pos + n)

            return b"".join(r)

    def readinto(self, b: bytearray) -> int:
        """Read bytes into a pre-allocated bytes-like object b.

        Returns an int representing the number of bytes read (0 for EOF).
        """
        buf = self.read(len(b))
        length = len(buf)
        b[:length] = buf
        return length

    def _read(self, offset: int, length: int) -> bytes:
        """Read method that backs this aligned stream."""
        raise NotImplementedError("_read needs to be implemented by subclass")

    def readall(self) -> bytes:
        """Read until end of stream."""
        return self.read()

    def readoffset(self, offset: int, length: int) -> bytes:
        """Convenience method to read from a given offset.

        Args:
            offset: The offset in the stream to read from.
            length: The number of bytes to read.
        """
        self.seek(offset)
        return self.read(length)

    def peek(self, n: int) -> bytes:
        """Convenience method to peek from the current offset without advancing the stream position.

        Args:
            n: The number of bytes to peek.
        """
        pos = self._pos
        data = self.read(n)
        self._set_pos(pos)
        return data

    def close(self) -> None:
        """Close the stream. Does nothing by default."""


class RangeStream(AlignedStream):
    """Create a stream with a specific range from another file-like object.

    ASCII representation::

        Source file-like object
        |................................................|
                RangeStream with offset and size
                |............................|

    Args:
        fh: The source file-like object.
        offset: The offset the stream should start from on the source file-like object.
        size: The size the stream should be.
        align: The alignment size.
    """

    def __init__(self, fh: BinaryIO, offset: int, size: int | None, align: int = STREAM_BUFFER_SIZE):
        super().__init__(size, align)
        self._fh = fh
        self.offset = offset

    def _seek(self, pos: int, whence: int = io.SEEK_SET) -> int:
        if self.size is None and whence == io.SEEK_END:
            if (pos := self._fh.seek(pos, whence)) is None:
                pos = self._fh.tell()
            return max(0, pos - self.offset)
        return super()._seek(pos, whence)

    def _read(self, offset: int, length: int) -> bytes:
        read_length = min(length, self.size - offset) if self.size else length
        self._fh.seek(self.offset + offset)
        return self._fh.read(read_length)


class RelativeStream(RangeStream):
    """Create a relative stream from another file-like object.

    ASCII representation::

        Source file-like object
        |................................................|
                RelativeStream with offset
                |........................................|

    Args:
        fh: The source file-like object.
        offset: The offset the stream should start from on the source file-like object.
        size: Optional size the stream should be.
        align: The alignment size.
    """

    def __init__(self, fh: BinaryIO, offset: int, size: int | None = None, align: int = STREAM_BUFFER_SIZE):
        super().__init__(fh, offset, size, align)


class BufferedStream(RelativeStream):
    """Create a buffered stream from another file-like object.

    Optionally start from a specific offset.

    Args:
        fh: The source file-like object.
        offset: The offset the stream should start from.
        size: The size the stream should be.
        align: The alignment size.
    """

    def __init__(self, fh: BinaryIO, offset: int = 0, size: int | None = None, align: int = STREAM_BUFFER_SIZE):
        super().__init__(fh, offset, size, align)


class MappingStream(AlignedStream):
    """Create a stream from multiple mapped file-like objects.

    Args:
        size: The size the stream should be.
        align: The alignment size.
    """

    def __init__(self, size: int | None = None, align: int = STREAM_BUFFER_SIZE):
        super().__init__(size, align)
        self._runs: list[tuple[int, int, BinaryIO, int]] = []

    def add(self, offset: int, size: int, fh: BinaryIO, file_offset: int = 0) -> None:
        """Add a file-like object to the stream.

        Args:
            offset: The offset in the stream this fh maps to.
            size: The size that this mapped fh spans in the stream.
            fh: The file-like object to map.
            file_offset: The offset in the fh to start from.

        Note that there is no check on overlapping offsets and/or sizes.
        """
        self._runs.append((offset, size, fh, file_offset))
        self._runs = sorted(self._runs, key=lambda run: run[0])
        self._buf = None
        self.size = self._runs[-1][0] + self._runs[-1][1]

    def _get_run_idx(self, offset: int) -> tuple[int, int, BinaryIO, int]:
        """Find a mapping run for a given offset.

        Args:
            offset: The offset to find a mapping for.

        Returns:
            The run tuple if found.

        Raises:
            IOError: If no mapping is found for the given offset.
        """
        for idx, run in enumerate(self._runs):
            if run[0] <= offset < run[0] + run[1]:
                return idx

        raise EOFError(f"No mapping for offset {offset}")

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        run_idx = self._get_run_idx(offset)
        runlist_len = len(self._runs)
        size = self.size

        while length > 0:
            if run_idx >= runlist_len:
                # We somehow requested more data than we have runs for
                break

            run_offset, run_size, run_fh, run_file_offset = self._runs[run_idx]

            if run_offset > offset:
                # We landed in a gap, stop reading
                break

            run_pos = offset - run_offset
            run_remaining = run_size - run_pos

            if run_remaining < 0:
                break

            read_count = min(size - offset, min(run_remaining, length))

            run_fh.seek(run_file_offset + run_pos)
            result.append(run_fh.read(read_count))

            offset += read_count
            length -= read_count
            run_idx += 1

        return b"".join(result)


class RunlistStream(AlignedStream):
    """Create a stream from multiple runs on another file-like object.

    This is common in filesystems, where file data information is stored in "runs".
    A run is a ``(block_offset, block_count)`` tuple, meaning the amount of consecutive blocks from a
    specific starting block. A block_offset of ``None`` represents a sparse run, meaning it must simply
    return all ``\\x00`` bytes.

    Args:
        fh: The source file-like object.
        runlist: The runlist for this stream in block units.
        size: The size of the stream. This can be smaller than the total sum of blocks (to account for slack space).
        block_size: The block size in bytes.
        align: Optional alignment that differs from the block size, otherwise ``block_size`` is used as alignment.
    """

    def __init__(
        self, fh: BinaryIO, runlist: list[tuple[int, int]], size: int, block_size: int, align: int | None = None
    ):
        super().__init__(size, align or block_size)

        if isinstance(fh, RunlistStream):
            self._fh = fh._fh
        else:
            self._fh = fh

        self._runlist = []
        self._runlist_offsets = []

        self.runlist = runlist
        self.block_size = block_size

    @property
    def runlist(self) -> list[tuple[int, int]]:
        return self._runlist

    @runlist.setter
    def runlist(self, runlist: list[tuple[int, int]]) -> None:
        self._runlist = runlist
        self._runlist_offsets = []

        offset = 0
        # Create a list of starting offsets for each run so we can bisect that quickly when reading
        for _, block_count in self._runlist:
            if offset != 0:
                self._runlist_offsets.append(offset)
            offset += block_count

        self._buf = None

    def _read(self, offset: int, length: int) -> bytes:
        r = []

        block_offset = offset // self.block_size

        run_idx = bisect_right(self._runlist_offsets, block_offset)
        runlist_len = len(self.runlist)
        size = self.size

        while length > 0:
            if run_idx >= runlist_len:
                # We somehow requested more data than we have runs for
                break

            # If run_idx == 0, we only have a single run
            run_block_pos = 0 if run_idx == 0 else self._runlist_offsets[run_idx - 1]
            run_block_offset, run_block_count = self.runlist[run_idx]

            run_size = run_block_count * self.block_size
            run_pos = offset - run_block_pos * self.block_size
            run_remaining = run_size - run_pos

            # Sometimes the self.size is way larger than what we actually have runs for?
            # Stop reading if we reach a negative run_remaining
            if run_remaining < 0:
                break

            read_count = min(size - offset, min(run_remaining, length))

            # Sparse run
            if run_block_offset is None:
                r.append(b"\x00" * read_count)
            else:
                self._fh.seek(run_block_offset * self.block_size + run_pos)
                r.append(self._fh.read(read_count))

            offset += read_count
            length -= read_count
            run_idx += 1

        return b"".join(r)


class OverlayStream(AlignedStream):
    """Create a stream from another file-like object with the ability to overlay other streams or bytes.

    Useful for patching large file-like objects without having to cache the entire contents.
    First wrap the original stream in this class, and then call ``add()`` with the offset and data to overlay.

    Args:
        fh: The source file-like object.
        size: The size the stream should be.
        align: The alignment size.
    """

    def __init__(self, fh: BinaryIO, size: int | None = None, align: int = STREAM_BUFFER_SIZE):
        super().__init__(size, align)
        self._fh = fh
        self.overlays: dict[int, tuple[int, BinaryIO]] = {}
        self._lookup: list[int] = []

    def add(self, offset: int, data: bytes | BinaryIO, size: int | None = None) -> None:
        """Add an overlay at the given offset.

        Args:
            offset: The offset in bytes to add an overlay at.
            data: The bytes or file-like object to overlay.
            size: Optional size specification of the overlay, if it can't be inferred.
        """
        if not hasattr(data, "read"):
            size = size or len(data)
            data = io.BytesIO(data)
        elif size is None:
            size = data.size if hasattr(data, "size") else data.seek(0, io.SEEK_END)

        if not size:
            return None

        if size < 0:
            raise ValueError("Size must be positive")

        # Check if there are overlapping overlays
        for other_offset, (other_size, _) in self.overlays.items():
            if other_offset < offset + size and offset < other_offset + other_size:
                raise ValueError(f"Overlap with existing overlay: ({other_offset, other_size})")

        self.overlays[offset] = (size, data)
        self._lookup.append(offset)
        self._lookup.sort()

        # Clear the buffer if we add an overlay at our current position
        if self._buf and (self._pos_align <= offset + size and offset <= self._pos_align + len(self._buf)):
            self._buf = None

        return self

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        fh = self._fh
        overlays = self.overlays
        lookup = self._lookup

        overlay_len = len(overlays)
        overlay_idx = bisect_left(lookup, offset)

        while length > 0:
            prev_overlay_offset = None if overlay_idx == 0 else lookup[overlay_idx - 1]
            next_overlay_offset = None if overlay_idx >= overlay_len else lookup[overlay_idx]

            if prev_overlay_offset is not None:
                prev_overlay_size, prev_overlay_data = overlays[prev_overlay_offset]
                prev_overlay_end = prev_overlay_offset + prev_overlay_size

                if prev_overlay_end > offset:
                    # Currently in an overlay
                    offset_in_prev_overlay = offset - prev_overlay_offset
                    prev_overlay_remaining = prev_overlay_size - offset_in_prev_overlay
                    prev_overlay_read_size = min(length, prev_overlay_remaining)

                    prev_overlay_data.seek(offset_in_prev_overlay)
                    result.append(prev_overlay_data.read(prev_overlay_read_size))

                    offset += prev_overlay_read_size
                    length -= prev_overlay_read_size

            if length == 0:
                break

            if next_overlay_offset is not None:
                next_overlay_size, next_overlay_data = overlays[next_overlay_offset]
                gap_to_next_overlay = next_overlay_offset - offset

                if 0 <= gap_to_next_overlay < length:
                    if gap_to_next_overlay:
                        fh.seek(offset)
                        result.append(fh.read(gap_to_next_overlay))

                    # read remaining from overlay
                    next_overlay_read_size = min(next_overlay_size, length - gap_to_next_overlay)
                    next_overlay_data.seek(0)
                    result.append(next_overlay_data.read(next_overlay_read_size))

                    offset += next_overlay_read_size + gap_to_next_overlay
                    length -= next_overlay_read_size + gap_to_next_overlay
                else:
                    # Next overlay is too far away, complete read
                    fh.seek(offset)
                    result.append(fh.read(length))
                    break
            else:
                # No next overlay, complete read
                fh.seek(offset)
                result.append(fh.read(length))
                break

            overlay_idx += 1

        return b"".join(result)


class ZlibStream(AlignedStream):
    """Create a zlib stream from another file-like object.

    Basically the same as ``gzip.GzipFile`` but for raw zlib streams.
    Due to the nature of zlib streams, seeking backwards requires resetting the decompression context.

    Args:
        fh: The source file-like object.
        size: The size the stream should be.
    """

    def __init__(self, fh: BinaryIO, size: int | None = None, align: int = STREAM_BUFFER_SIZE, **kwargs):
        self._fh = fh

        self._zlib = None
        self._zlib_args = kwargs
        self._zlib_offset = 0
        self._zlib_prepend = b""
        self._zlib_prepend_offset = None
        self._rewind()

        super().__init__(size, align)

    def _rewind(self) -> None:
        self._fh.seek(0)
        self._zlib = zlib.decompressobj(**self._zlib_args)
        self._zlib_offset = 0
        self._zlib_prepend = b""
        self._zlib_prepend_offset = None

    def _seek_zlib(self, offset: int) -> None:
        if offset < self._zlib_offset:
            self._rewind()

        while self._zlib_offset < offset:
            read_size = min(offset - self._zlib_offset, self.align)
            if self._read_zlib(read_size) == b"":
                break

    def _read_fh(self, length: int) -> bytes:
        if self._zlib_prepend_offset is None:
            return self._fh.read(length)

        if self._zlib_prepend_offset + length <= len(self._zlib_prepend):
            offset = self._zlib_prepend_offset
            self._zlib_prepend_offset += length
            return self._zlib_prepend[offset : self._zlib_prepend_offset]

        offset = self._zlib_prepend_offset
        self._zlib_prepend_offset = None
        return self._zlib_prepend[offset:] + self._fh.read(length - len(self._zlib_prepend) + offset)

    def _read_zlib(self, length: int) -> bytes:
        if length < 0:
            return self.readall()

        result = []
        while length > 0:
            buf = self._read_fh(io.DEFAULT_BUFFER_SIZE)
            decompressed = self._zlib.decompress(buf, length)

            if self._zlib.unconsumed_tail != b"":
                self._zlib_prepend = self._zlib.unconsumed_tail
                self._zlib_prepend_offset = 0

            if buf == b"":
                break

            result.append(decompressed)
            length -= len(decompressed)

        buf = b"".join(result)
        self._zlib_offset += len(buf)
        return buf

    def _read(self, offset: int, length: int) -> bytes:
        self._seek_zlib(offset)
        return self._read_zlib(length)

    def readall(self) -> bytes:
        self._seek_zlib(self.tell())

        chunks = []
        # sys.maxsize means the max length of output buffer is unlimited,
        # so that the whole input buffer can be decompressed within one
        # .decompress() call.
        while data := self._read_zlib(sys.maxsize):
            chunks.append(data)

        return b"".join(chunks)


class BitStream:
    """Bit-level stream reader.

    Args:
        fh: File-like object to read bits from.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self._byte_offset = fh.tell()

        self.buffer = 0
        self.bits = 0

    def readable(self) -> bool:
        """Indicate that the stream is readable."""
        return True

    def seekable(self) -> bool:
        """Indicate that the stream is seekable."""
        return True

    def writable(self) -> bool:
        """Indicate that the stream is not writable."""
        return False

    def seek(self, pos: int, whence: int = io.SEEK_SET) -> int:
        """Seek the stream to the specified position in bits.

        Returns:
            The new stream position after seeking.
        """
        if whence == io.SEEK_SET:
            byte_pos, bit_pos = divmod(pos, 8)
        elif whence == io.SEEK_CUR:
            current_pos = self.tell()
            byte_pos, bit_pos = divmod(current_pos + pos, 8)
        elif whence == io.SEEK_END:
            self.fh.seek(0, io.SEEK_END)
            end_pos = self.fh.tell() * 8
            byte_pos, bit_pos = divmod(end_pos + pos, 8)
        else:
            raise IOError("invalid whence value")

        self._byte_offset = byte_pos
        self.bits = 0
        self.buffer = 0

        if bit_pos > 0:
            self.read(bit_pos)

        return self.tell()

    def tell(self) -> int:
        """Get the current position in the stream in bits."""
        return (self._byte_offset * 8) - self.bits

    def read(self, n: int) -> int:
        """Read n bits from the stream.

        Args:
            n: Number of bits to read.
        """
        value = self.peek(n)
        self.remove(n)
        return value

    def peek(self, n: int) -> int:
        """Peek n bits from the stream without advancing.

        Args:
            n: Number of bits to peek.
        """
        if n == 0:
            return 0

        if n > self.bits:
            while self.bits < n:
                num_bytes = (n - self.bits + 7) // 8
                self.fh.seek(self._byte_offset)
                if not (buf := self.fh.read(num_bytes)):
                    break

                new_bits = int.from_bytes(buf, "big")
                num_new_bits = len(buf) * 8
                self.buffer = (self.buffer << num_new_bits) | new_bits
                self.bits += num_new_bits
                self._byte_offset += len(buf)

        return self.buffer >> (self.bits - min(n, self.bits))

    def remove(self, n: int) -> None:
        """Remove n bits from the stream.

        Args:
            n: Number of bits to remove.
        """
        self.bits -= min(n, self.bits)
        self.buffer &= (1 << (self.bits)) - 1



from __future__ import annotations

import struct
from datetime import datetime, timedelta, timezone

# Python on Windows and WASM (Emscripten) have problems calculating timestamps before 1970 (Unix epoch)
# Calculating relatively from the epoch is required on these platforms
# This method is slower, so we split the implementation between Windows, WASM and other platforms
# This used to be a platform comparison, but that was not reliable enough, so ducktype it instead
try:
    datetime.fromtimestamp(-6969696969, tz=timezone.utc)

    def _calculate_timestamp(ts: float) -> datetime:
        """Calculate timestamps normally."""
        return datetime.fromtimestamp(ts, tz=timezone.utc)
except (OSError, OverflowError):
    _EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)

    def _calculate_timestamp(ts: float) -> datetime:
        """Calculate timestamps relative from Unix epoch."""
        return _EPOCH + timedelta(seconds=ts)


def now() -> datetime:
    """Return an aware datetime object of the current time in UTC."""
    return datetime.now(timezone.utc)


def unix_now() -> int:
    """Return a Unix timestamp of the current time."""
    return to_unix(now())


def unix_now_ms() -> int:
    """Return a Unix millisecond timestamp of the current time."""
    return to_unix_ms(now())


def unix_now_us() -> int:
    """Return a Unix microsecond timestamp of the current time."""
    return to_unix_us(now())


def unix_now_ns() -> int:
    """Return a Unix nanosecond timestamp of the current time."""
    return to_unix_ns(now())


def to_unix(dt: datetime) -> int:
    """Converts datetime objects into Unix timestamps.

    This is a convenience method.

    Args:
        dt: The datetime object.

    Returns:
        Unix timestamp from the passed datetime object.
    """
    return int(dt.timestamp())


def to_unix_ms(dt: datetime) -> int:
    """Converts datetime objects into Unix millisecond timestamps.

    This is a convenience method.

    Args:
        dt: The datetime object.

    Returns:
        Unix millisecond timestamp from the passed datetime object.
    """
    return int(dt.timestamp() * 1e3)


def to_unix_us(dt: datetime) -> int:
    """Converts datetime objects into Unix microsecond timestamps.

    This is a convenience method.

    Args:
        dt: The datetime object.

    Returns:
        Unix microsecond timestamp from the passed datetime object.
    """
    return int(dt.timestamp() * 1e6)


def to_unix_ns(dt: datetime) -> int:
    """Converts datetime objects into Unix nanosecond timestamps.

    This is a convenience method.

    Args:
        dt: The datetime object.

    Returns:
        Unix nanosecond timestamp from the passed datetime object.
    """
    return to_unix_us(dt) * 1000


def from_unix(ts: float) -> datetime:
    """Converts Unix timestamps to aware datetime objects in UTC.

    This is a convenience method.

    Args:
        ts: The Unix timestamp.

    Returns:
        Datetime object from the passed timestamp.
    """
    return _calculate_timestamp(ts)


def from_unix_ms(ts: float) -> datetime:
    """Converts Unix timestamps in milliseconds to aware datetime objects in UTC.

    Args:
        ts: The Unix timestamp in milliseconds.

    Returns:
        Datetime object from the passed timestamp.
    """
    return from_unix(float(ts) * 1e-3)


def from_unix_us(ts: float) -> datetime:
    """Converts Unix timestamps in microseconds to aware datetime objects in UTC.

    Args:
        ts: The Unix timestamp in microseconds.

    Returns:
        Datetime object from the passed timestamp.
    """
    return from_unix(float(ts) * 1e-6)


def from_unix_ns(ts: float) -> datetime:
    """Converts Unix timestamps in nanoseconds to aware datetime objects in UTC.

    Args:
        ts: The Unix timestamp in nanoseconds.

    Returns:
        Datetime object from the passed timestamp.
    """
    return from_unix(float(ts) * 1e-9)


def xfstimestamp(seconds: int, nano: int) -> datetime:
    """Converts XFS timestamps to aware datetime objects in UTC.

    Args:
        seconds: The XFS timestamp seconds component
        nano: The XFS timestamp nano seconds component
    Returns:
        Datetime object from the passed timestamp.
    """
    return _calculate_timestamp(float(seconds) + (1e-9 * nano))


ufstimestamp = xfstimestamp


def wintimestamp(ts: int | tuple[int, int]) -> datetime:
    """Converts Windows ``FILETIME`` timestamps to aware datetime objects in UTC.

    Args:
        ts: The Windows timestamp integer or a tuple of integers (``dwLowDateTime``, ``dwHighDateTime``)

    Returns:
        Datetime object from the passed timestamp.

    Resources:
        - https://learn.microsoft.com/en-us/windows/win32/api/minwinbase/ns-minwinbase-filetime
    """
    if isinstance(ts, tuple):
        if len(ts) != 2:
            raise ValueError(f"Expected (dwLowDateTime, dwHighDateTime) tuple but got {ts!r}")
        ts = (ts[1] << 32) + ts[0]

    return _calculate_timestamp(float(ts) * 1e-7 - 11_644_473_600)  # Thanks FireEye


def oatimestamp(ts: float) -> datetime:
    """Converts OLE Automation timestamps to aware datetime objects in UTC.

    Args:
        ts: The OLE Automation timestamp.

    Returns:
        Datetime object from the passed timestamp.
    """
    if not isinstance(ts, float):
        # Convert from int to float
        (ts,) = struct.unpack("<d", struct.pack("<Q", ts & 0xFFFFFFFFFFFFFFFF))
    return _calculate_timestamp((ts - 25569) * 86400)


def webkittimestamp(ts: int) -> datetime:
    """Converts WebKit timestamps to aware datetime objects in UTC.

    Args:
        ts: The WebKit timestamp.

    Returns:
        Datetime object from the passed timestamp.
    """
    return _calculate_timestamp(float(ts) * 1e-6 - 11644473600)


def cocoatimestamp(ts: int) -> datetime:
    """Converts Apple Cocoa Core Data timestamps to aware datetime objects in UTC.

    Args:
        ts: The Apple Cocoa Core Data timestamp.

    Returns:
        Datetime object from the passed timestamp.
    """
    return _calculate_timestamp(float(ts) + 978307200)


def uuid1timestamp(ts: int) -> datetime:
    """Converts UUID version 1 timestamps to aware datetime objects in UTC.

    UUID v1 timestamps have an epoch of 1582-10-15 00:00:00.

    Args:
        ts: The UUID version 1 timestamp

    Returns:
        Datetime object from the passed timestamp.
    """
    return _calculate_timestamp(float(ts) * 1e-7 - 12219292800)


DOS_EPOCH_YEAR = 1980


def dostimestamp(ts: int, centiseconds: int = 0, swap: bool = False) -> datetime:
    """Converts MS-DOS timestamps to naive datetime objects.

    MS-DOS timestamps are recorded in local time, so we leave it up to the caller to add optional timezone information.

    References:
        - https://web.archive.org/web/20180311003959/http://www.vsft.com/hal/dostime.htm

    Args:
        ts: MS-DOS timestamp
        centiseconds: Optional ExFAT centisecond offset. Yes centisecond...
        swap: Optional swap flag if date and time bytes are swapped.

    Returns:
        Datetime object from the passed timestamp.
    """
    # MS-DOS Date Time Format is actually 2 UINT16_T's first 16 bits are the time, second 16 bits are date
    # the year is an offset of the MS-DOS epoch year, which is 1980

    if swap:
        year = ((ts >> 9) & 0x7F) + DOS_EPOCH_YEAR
        month = (ts >> 5) & 0x0F
        day = ts & 0x1F

        hours = (ts >> 27) & 0x1F
        minutes = (ts >> 21) & 0x3F
        seconds = ((ts >> 16) & 0x1F) * 2
    else:  # non-swapped way
        year = ((ts >> 25) & 0x7F) + DOS_EPOCH_YEAR
        month = (ts >> 21) & 0x0F
        day = (ts >> 16) & 0x1F

        hours = (ts >> 11) & 0x1F
        minutes = (ts >> 5) & 0x3F
        seconds = (ts & 0x1F) * 2

    # Note that according to the standard, centiseconds can be at most 199, so
    # extra_seconds will be at most 1.
    extra_seconds, centiseconds = divmod(centiseconds, 100)
    microseconds = centiseconds * 10000

    return datetime(  # noqa: DTZ001
        year,
        month or 1,
        day or 1,
        hours,
        minutes,
        seconds + extra_seconds,
        microseconds,
    )



from __future__ import annotations

from io import BytesIO
from typing import BinaryIO


def compress(src: bytes | BinaryIO) -> bytes:
    """Sevenbit compress from a file-like object or bytes.

    Args:
        src: File-like object or bytes to compress.

    Returns:
        The compressed data.
    """
    if not hasattr(src, "read"):
        src = BytesIO(src)

    dst = bytearray()

    val = 0
    shift = 0
    while True:
        _byte = src.read(1)
        if not len(_byte):
            break

        val |= (_byte[0] & 0x7F) << shift
        shift += 7

        if shift >= 8:
            dst.append(val & 0xFF)
            val >>= 8
            shift -= 8

    if val:
        dst.append(val & 0xFF)

    return bytes(dst)


def decompress(src: bytes | BinaryIO, wide: bool = False) -> bytes:
    """Sevenbit decompress from a file-like object or bytes.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = BytesIO(src)

    dst = bytearray()

    val = 0
    shift = 0
    while True:
        _byte = src.read(1)
        if not len(_byte):
            break

        val |= _byte[0] << shift
        dst.append(val & 0x7F)
        if wide:
            dst.append(0)

        val >>= 7
        shift += 1
        if shift == 7:
            dst.append(val & 0x7F)
            if wide:
                dst.append(0)
            val >>= 7
            shift = 0

    return bytes(dst)



from __future__ import annotations

import zlib
from functools import lru_cache


@lru_cache(maxsize=32)
def _table(polynomial: int) -> tuple[int, ...]:
    """Generate a CRC32 table for a given (reversed) polynomial.

    Args:
        polynomial: The (reversed) polynomial to use for the CRC32 calculation.
    """
    table = []
    for i in range(256):
        crc = i
        for _ in range(8):
            if (crc & 1) != 0:
                crc = (crc >> 1) ^ polynomial
            else:
                crc >>= 1
            crc &= 0xFFFFFFFF
        table.append(crc)
    return tuple(table)


def update(crc: int, data: bytes, polynomial: int = 0xEDB88320, table: tuple[int, ...] | None = None) -> int:
    """Update CRC32 checksum with data.

    Args:
        crc: The initial value of the checksum.
        data: The data to update the checksum with.
        polynomial: The (reversed) polynomial to use for the CRC32 calculation. Default is 0xEDB88320 (crc32b).
        table: Optional precomputed CRC32 table. If None, a table will be generated using the given polynomial.
    """
    if polynomial == 0xEDB88320 and table is None:
        return zlib.crc32(data, crc)

    if table is None:
        table = _table(polynomial)

    crc = crc ^ 0xFFFFFFFF
    for b in data:
        crc = table[(crc ^ b) & 0xFF] ^ ((crc >> 8) & 0xFFFFFFFF)
    return crc ^ 0xFFFFFFFF


def crc32(data: bytes, value: int = 0, polynomial: int = 0xEDB88320, table: tuple[int, ...] | None = None) -> int:
    """Calculate CRC32 checksum of some data, with an optional initial value and polynomial.

    Args:
        data: The data to calculate the checksum of.
        value: The initial value of the checksum. Default is 0.
        polynomial: The (reversed) polynomial to use for the CRC32 calculation. Default is 0xEDB88320 (crc32b).
        table: Optional precomputed CRC32 table. If None, a table will be generated using the given polynomial.
    """
    return update(value, data, polynomial, table) & 0xFFFFFFFF



from __future__ import annotations

import stat
import struct
import tarfile
from tarfile import InvalidHeaderError
from typing import BinaryIO

FORMAT_CPIO_BIN = 10
FORMAT_CPIO_ODC = 11
FORMAT_CPIO_NEWC = 12
FORMAT_CPIO_CRC = 13
FORMAT_CPIO_HPBIN = 16
FORMAT_CPIO_HPODC = 17
FORMAT_CPIO_UNKNOWN = 18

CPIO_MAGIC_OLD = 0o070707
CPIO_MAGIC_NEW = 0o070701
CPIO_MAGIC_CRC = 0o070702

TYPE_MAP = {
    stat.S_IFREG: tarfile.REGTYPE,
    stat.S_IFDIR: tarfile.DIRTYPE,
    stat.S_IFIFO: tarfile.FIFOTYPE,
    stat.S_IFLNK: tarfile.SYMTYPE,
    stat.S_IFCHR: tarfile.CHRTYPE,
    stat.S_IFBLK: tarfile.BLKTYPE,
}


class CpioInfo(tarfile.TarInfo):
    """Custom ``TarInfo`` implementation for reading cpio archives.

    Examples::

        tarfile.open(..., tarinfo=CpioInfo)
        # or
        tarfile.TarFile(..., tarinfo=CpioInfo)

    """

    @classmethod
    def fromtarfile(cls, tarfile: tarfile.TarFile) -> CpioInfo:
        if tarfile.format not in (
            FORMAT_CPIO_BIN,
            FORMAT_CPIO_ODC,
            FORMAT_CPIO_NEWC,
            FORMAT_CPIO_CRC,
            FORMAT_CPIO_HPBIN,
            FORMAT_CPIO_HPODC,
        ):
            tarfile.format = detect_header(tarfile.fileobj)

        if tarfile.format in (FORMAT_CPIO_BIN, FORMAT_CPIO_HPBIN):
            buf = tarfile.fileobj.read(26)
        elif tarfile.format in (FORMAT_CPIO_ODC, FORMAT_CPIO_HPODC):
            buf = tarfile.fileobj.read(76)
        elif tarfile.format in (FORMAT_CPIO_NEWC, FORMAT_CPIO_CRC):
            buf = tarfile.fileobj.read(110)
        else:
            raise InvalidHeaderError("Unknown cpio type")

        obj = cls.frombuf(buf, tarfile.format, tarfile.encoding, tarfile.errors)
        obj.format = tarfile.format
        obj.offset = tarfile.fileobj.tell() - len(buf)
        return obj._proc_member(tarfile)

    @classmethod
    def frombuf(cls, buf: bytes, format: int, encoding: str, errors: str) -> CpioInfo:
        if format in (FORMAT_CPIO_BIN, FORMAT_CPIO_ODC, FORMAT_CPIO_HPBIN, FORMAT_CPIO_HPODC):
            obj = cls._old_frombuf(buf, format)
        elif format in (FORMAT_CPIO_NEWC, FORMAT_CPIO_CRC):
            obj = cls._new_frombuf(buf, format)

        # Common postprocessing
        ftype = stat.S_IFMT(obj._mode)
        obj.type = TYPE_MAP.get(ftype, ftype)
        obj.mode = stat.S_IMODE(obj._mode)

        return obj

    @classmethod
    def _old_frombuf(cls, buf: bytes, format: int) -> CpioInfo:
        if format in (FORMAT_CPIO_BIN, FORMAT_CPIO_HPBIN):
            values = list(struct.unpack("<13H", buf))
            if values[0] == _swap16(CPIO_MAGIC_OLD):
                values = [_swap16(v) for v in values]

            mtime = (values.pop(8) << 16) | values.pop(8)
            size = (values.pop(9) << 16) | values.pop(9)
            values.insert(8, mtime)
            values.append(size)
        else:
            values = [int(v, 8) for v in struct.unpack("<6s6s6s6s6s6s6s6s11s6s11s", buf)]

        if values[0] != CPIO_MAGIC_OLD:
            raise InvalidHeaderError(f"Invalid (old) ASCII/binary cpio header magic: {oct(values[0])}")

        obj = cls()
        obj.devmajor = values[1] >> 8
        obj.devminor = values[1] & 0xFF
        obj._mode = values[3]
        obj.uid = values[4]
        obj.gid = values[5]
        obj.mtime = values[8]
        obj.size = values[10]

        # Extra fields
        obj.magic = values[0]
        obj.ino = values[2]
        obj.nlink = values[6]
        obj.rdevmajor = values[7] >> 8
        obj.rdevminor = values[7] & 0xFF
        obj.namesize = values[9]

        # This is a specific case for HP/UX cpio archives, which I'll let this comment from the original source explain:
        # HP/UX cpio creates archives that look just like ordinary archives,
        # but for devices it sets major = 0, minor = 1, and puts the
        # actual major/minor number in the filesize field.  See if this
        # is an HP/UX cpio archive, and if so fix it.  We have to do this
        # here because process_copy_in() assumes filesize is always 0
        # for devices.
        if (
            stat.S_IFMT(obj.mode) in (stat.S_IFCHR, stat.S_IFBLK, stat.S_IFSOCK, stat.S_IFIFO)
            and obj.size != 0
            and obj.rdevmajor == 0
            and obj.rdevminor == 1
        ):
            obj.rdevmajor = (obj.size >> 8) & 0xFF
            obj.rdevminor = obj.size & 0xFF
            obj.size = 0

        return obj

    @classmethod
    def _new_frombuf(cls, buf: bytes, format: int) -> CpioInfo:
        values = struct.unpack("<6s8s8s8s8s8s8s8s8s8s8s8s8s8s", buf)
        values = [int(values[0], 8)] + [int(v, 16) for v in values[1:]]
        if values[0] not in (CPIO_MAGIC_NEW, CPIO_MAGIC_CRC):
            raise InvalidHeaderError(f"Invalid (new) ASCII cpio header magic: {oct(values[0])}")

        obj = cls()
        obj._mode = values[2]
        obj.uid = values[3]
        obj.gid = values[4]
        obj.mtime = values[6]
        obj.size = values[7]
        obj.devmajor = values[8]
        obj.devminor = values[9]
        obj.chksum = values[13]

        # Extra fields
        obj.magic = values[0]
        obj.ino = values[1]
        obj.nlink = values[5]
        obj.rdevmajor = values[10]
        obj.rdevminor = values[11]
        obj.namesize = values[12]

        return obj

    def _proc_member(self, tarfile: tarfile.TarFile) -> CpioInfo | None:
        self.name = tarfile.fileobj.read(self.namesize - 1).decode(tarfile.encoding, tarfile.errors)
        if self.name == "TRAILER!!!":
            # The last entry in a cpio file has the special name ``TRAILER!!!``, indicating the end of the archive
            return None

        offset = tarfile.fileobj.tell() + 1
        self.offset_data = self._round_word(offset)
        tarfile.offset = self._round_word(self.offset_data + self.size)

        if self.issym():
            tarfile.fileobj.seek(self.offset_data)
            self.linkname = tarfile.fileobj.read(self.size).decode(tarfile.encoding, tarfile.errors)
            self.size = 0

        return self

    def _round_word(self, offset: int) -> int:
        if self.format in (FORMAT_CPIO_BIN, FORMAT_CPIO_HPBIN):
            return (offset + 1) & ~0x01

        if self.format in (FORMAT_CPIO_NEWC, FORMAT_CPIO_CRC):
            return (offset + 3) & ~0x03

        return offset

    def issocket(self) -> bool:
        """Return True if it is a socket."""
        return self.type == stat.S_IFSOCK


def detect_header(fh: BinaryIO) -> int:
    """Detect a cpio format on a file-like object."""
    offset = fh.tell()
    magic = fh.read(6)
    fh.seek(offset)

    result = FORMAT_CPIO_UNKNOWN
    if magic == b"070701":
        result = FORMAT_CPIO_NEWC
    elif magic == b"070707":
        result = FORMAT_CPIO_ODC
    elif magic == b"070702":
        result = FORMAT_CPIO_CRC
    elif magic[:2] in (b"\x71\xc7", b"\xc7\x71"):
        # 0o070707 in little and big endian
        result = FORMAT_CPIO_BIN

    return result


def _swap16(value: int) -> int:
    return ((value & 0xFF) << 8) | (value >> 8)


def CpioFile(*args, **kwargs) -> tarfile.TarFile:
    """Utility wrapper around ``tarfile.TarFile`` to easily open cpio archives."""
    kwargs.setdefault("format", FORMAT_CPIO_UNKNOWN)
    return tarfile.TarFile(*args, **kwargs, tarinfo=CpioInfo)


def open(*args, **kwargs) -> tarfile.TarFile:
    """Utility wrapper around ``tarfile.open`` to easily open cpio archives."""
    kwargs.setdefault("format", FORMAT_CPIO_UNKNOWN)
    return tarfile.open(*args, **kwargs, tarinfo=CpioInfo)



from __future__ import annotations

import plistlib
import uuid
from collections import UserDict
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.util.ts import cocoatimestamp

if TYPE_CHECKING:
    from datetime import datetime


class NSKeyedArchiver:
    def __init__(self, fh: BinaryIO):
        self.plist = plistlib.load(fh)

        if not isinstance(self.plist, dict) or not all(
            key in self.plist for key in ["$version", "$archiver", "$top", "$objects"]
        ):
            raise ValueError("File is not an NSKeyedArchiver plist")

        self._objects = self.plist.get("$objects")
        self._cache = {}

        self.top = {}
        for name, value in self.plist.get("$top", {}).items():
            self.top[name] = self._parse(value)

    def __getitem__(self, key: str) -> Any:
        return self.top[key]

    def __repr__(self) -> str:
        return f"<NSKeyedArchiver top={self.top}>"

    def get(self, key: str, default: Any | None = None) -> Any:
        return self.top.get(key, default)

    def _parse(self, uid: Any) -> Any:
        if not isinstance(uid, plistlib.UID):
            return uid

        num = uid.data
        if num in self._cache:
            return self._cache[num]
        result = self._parse_obj(self._objects[num])
        self._cache[num] = result
        return result

    def _parse_obj(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            klass = obj.get("$class")
            if klass:
                klass_name = self._parse(klass).get("$classname")
                return CLASSES.get(klass_name, NSObject)(self, obj)
            return obj

        if isinstance(obj, list):
            return list(map(self._parse, obj))

        if isinstance(obj, bool | bytes | int | float) or obj is None:
            return obj

        if isinstance(obj, str):
            return None if obj == "$null" else obj

        return None


class NSObject:
    def __init__(self, nskeyed: NSKeyedArchiver, obj: dict[str, Any]):
        self.nskeyed = nskeyed
        self.obj = obj

        self._class = nskeyed._parse(obj.get("$class", {}))
        self._classname = self._class.get("$classname", "Unknown")
        self._classes = self._class.get("$classes", [])

    def __getitem__(self, attr: str) -> Any:
        obj = self.obj[attr]
        return self.nskeyed._parse(obj)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __repr__(self):
        return f"<{self._classname}>"

    def keys(self) -> list[str]:
        return self.obj.keys()

    def get(self, attr: str, default: Any | None = None) -> Any:
        try:
            return self[attr]
        except KeyError:
            return default


class NSDictionary(UserDict, NSObject):
    def __init__(self, nskeyed: NSKeyedArchiver, obj: dict[str, Any]):
        NSObject.__init__(self, nskeyed, obj)
        self.data = {nskeyed._parse(key): obj for key, obj in zip(obj["NS.keys"], obj["NS.objects"], strict=False)}

    def __repr__(self) -> str:
        return NSObject.__repr__(self)

    def __getitem__(self, key: str) -> Any:
        return self.nskeyed._parse(self.data[key])


def parse_nsarray(nskeyed: NSKeyedArchiver, obj: dict[str, Any]) -> list[Any]:
    return list(map(nskeyed._parse, obj["NS.objects"]))


def parse_nsset(nskeyed: NSKeyedArchiver, obj: dict[str, Any]) -> list[Any]:
    # Some values are not hashable, so return as list
    return parse_nsarray(nskeyed, obj)


def parse_nsdata(nskeyed: NSKeyedArchiver, obj: dict[str, Any]) -> Any:
    return obj["NS.data"]


def parse_nsdate(nskeyed: NSKeyedArchiver, obj: dict[str, Any]) -> datetime:
    return cocoatimestamp(obj["NS.time"])


def parse_nsuuid(nskeyed: NSKeyedArchiver, obj: dict[str, Any]) -> uuid.UUID:
    return uuid.UUID(bytes=obj["NS.uuidbytes"])


def parse_nsurl(nskeyed: NSKeyedArchiver, obj: dict[str, Any]) -> str:
    base = nskeyed._parse(obj["NS.base"])
    relative = nskeyed._parse(obj["NS.relative"])
    if base:
        return f"{base}/{relative}"
    return relative


CLASSES = {
    "NSArray": parse_nsarray,
    "NSMutableArray": parse_nsarray,
    "NSDictionary": NSDictionary,
    "NSMutableDictionary": NSDictionary,
    "NSSet": parse_nsset,
    "NSMutableSet": parse_nsset,
    "NSData": parse_nsdata,
    "NSMutableData": parse_nsdata,
    "NSDate": parse_nsdate,
    "NSUUID": parse_nsuuid,
    "NSURL": parse_nsurl,
    "NSNull": lambda nskeyed, obj: None,
}



# References:
# - https://github.com/eafer/libzbitmap
# - https://github.com/sgan81/apfs-fuse
from __future__ import annotations

import io
import itertools
import struct
from typing import BinaryIO

_H = struct.Struct("<H")


def decompress(src: bytes | BinaryIO) -> bytes:
    """LZBITMAP decompress from a file-like object or bytes.

    Decompresses until EOF or EOS of the input data.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    if src.read(4) != b"ZBM\x09":
        raise ValueError("Not a valid LZBITMAP stream")

    dst = bytearray()

    while True:
        compressed_size = int.from_bytes(src.read(3), "little")
        uncompressed_size = int.from_bytes(src.read(3), "little")

        if compressed_size == uncompressed_size + 6:  # chunk header size
            # Not compressed
            dst += src.read(uncompressed_size)

        elif uncompressed_size == 0:
            # End of stream
            break

        else:
            # Compressed
            distance_offset = int.from_bytes(src.read(3), "little")
            bitmap_offset = int.from_bytes(src.read(3), "little")
            token_offset = int.from_bytes(src.read(3), "little")
            literal_offset = 15

            # Buffer the whole chunk
            src.seek(-15, io.SEEK_CUR)
            buf = memoryview(src.read(compressed_size))

            # Build the bitmap/token map
            token_map = []
            bits = int.from_bytes(buf[-17:], "little")
            for i in range(0xF):
                if i < 3:
                    token_map.append((None, i))
                else:
                    token_map.append((bits & 0xFF, (bits >> 8) & 3))
                    bits >>= 10

            # Tokens are stored as nibbles, so we need to split each byte into two
            tokens = itertools.chain.from_iterable((b & 0xF, (b >> 4) & 0xF) for b in buf[token_offset:-17])

            # Initial match distance is 8, and is not reset between tokens
            distance = 8

            prev_token = None
            while uncompressed_size > 0:
                if (idx := next(tokens, None) if prev_token is None else prev_token) is None:
                    break

                if idx == 0xF:
                    # 0xF indicates a repeat count
                    raise ValueError("Invalid token index in LZBITMAP stream")

                if (repeat := next(tokens)) != 0xF:
                    # No repeat count, store the token for next iteration
                    prev_token = repeat
                    repeat = 1
                else:
                    # Repeat count, read and sum
                    prev_token = None
                    total = 4
                    while repeat == 0xF:
                        repeat = next(tokens)
                        total += repeat

                        if total < repeat:
                            raise ValueError("Invalid repeat count in LZBITMAP stream")

                    repeat = total

                for _ in range(repeat):
                    bitmap, token = token_map[idx]
                    if idx < 3:
                        # Index 0, 1, 2 are special and indicate we need to read a bitmap from the bitmap region
                        bitmap = buf[bitmap_offset]
                        bitmap_offset += 1

                    if token == 1:
                        # 1-byte distance
                        distance = buf[distance_offset]
                        distance_offset += 1
                    elif token == 2:
                        # 2-byte distance
                        (distance,) = _H.unpack_from(buf, distance_offset)
                        distance_offset += 2

                    for _ in range(8):
                        if bitmap & 1:
                            # Literal
                            dst.append(buf[literal_offset])
                            literal_offset += 1
                        else:
                            # Match
                            if distance > len(dst):
                                raise ValueError("Invalid match distance in LZBITMAP stream")
                            dst.append(dst[-distance])

                        bitmap >>= 1
                        uncompressed_size -= 1
                        if uncompressed_size == 0:
                            break

    return bytes(dst)



# References:
# - https://github.com/lzfse/lzfse
from __future__ import annotations

import io
import struct
from typing import BinaryIO

# fmt: off
OP_SML_D = (
      0,   1,   2,   3,   4,   5,   8,   9,  10,  11,  12,  13,  16,  17,  18,  19,
     20,  21,  24,  25,  26,  27,  28,  29,  32,  33,  34,  35,  36,  37,  40,  41,
     42,  43,  44,  45,  48,  49,  50,  51,  52,  53,  56,  57,  58,  59,  60,  61,
     64,  65,  66,  67,  68,  69,  72,  73,  74,  75,  76,  77,  80,  81,  82,  83,
     84,  85,  88,  89,  90,  91,  92,  93,  96,  97,  98,  99, 100, 101, 104, 105,
    106, 107, 108, 109, 128, 129, 130, 131, 132, 133, 136, 137, 138, 139, 140, 141,
    144, 145, 146, 147, 148, 149, 152, 153, 154, 155, 156, 157, 192, 193, 194, 195,
    196, 197, 200, 201, 202, 203, 204, 205,
)
OP_MED_D = (
    160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175,
    176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191,
)
OP_LRG_D = (
      7,  15,  23,  31,  39,  47,  55,  63,  71,  79,  87,  95, 103, 111, 135, 143,
    151, 159, 199, 207,
)
OP_PRE_D = (
     70,  78,  86,  94, 102, 110, 134, 142, 150, 158, 198, 206,
)
OP_SML_M = (
    241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255,
)
OP_LRG_M = (
    240,
)
OP_SML_L = (
    225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239,
)
OP_LRG_L = (
    224,
)
OP_NOP = (
     14,  22,
)
OP_EOS = (
    6,
)
OP_UDEF = (
     30,  38,  46,  54,  62, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121,
    122, 123, 124, 125, 126, 127, 208, 209, 210, 211, 212, 213, 214, 215, 216,
    217, 218, 219, 220, 221, 222, 223,
)
# fmt: on

_H = struct.Struct("<H")


def decompress(src: bytes | BinaryIO) -> bytes:
    """LZVN decompress from a file-like object or bytes.

    Decompresses until EOF or EOS of the input data.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    offset = src.tell()
    src.seek(0, io.SEEK_END)
    src_size = src.tell() - offset
    src.seek(offset)

    dst = bytearray()

    opc_len = 0

    L = None
    M = None
    D = 0

    while src_size > 0:
        opc = src.read(1)[0]

        if opc in OP_SML_D:
            # "small distance": This opcode has the structure LLMMMDDD DDDDDDDD LITERAL
            # where the length of literal (0-3 bytes) is encoded by the high 2 bits of
            # the first byte. We first extract the literal length so we know how long
            # the opcode is, then check that the source can hold both this opcode and
            # at least one byte of the next (because any valid input stream must be
            # terminated with an eos token).
            opc_len = 2
            L = _extract(opc, 8, 6, 2)
            M = _extract(opc, 8, 3, 3) + 3

            if src_size <= opc_len + L:
                break

            D = _extract(opc, 8, 0, 3) << 8 | src.read(1)[0]

        elif opc in OP_MED_D:
            # "medium distance": This is a minor variant of the "small distance"
            # encoding, where we will now use two extra bytes instead of one to encode
            # the restof the match length and distance. This allows an extra two bits
            # for the match length, and an extra three bits for the match distance. The
            # full structure of the opcode is 101LLMMM DDDDDDMM DDDDDDDD LITERAL.
            opc_len = 3
            L = _extract(opc, 8, 3, 2)

            if src_size <= opc_len + L:
                break

            (opc23,) = _H.unpack(src.read(2))
            M = (_extract(opc, 8, 0, 3) << 2 | _extract(opc23, 16, 0, 2)) + 3
            D = _extract(opc23, 16, 2, 14)

        elif opc in OP_LRG_D:
            # "large distance": This is another variant of the "small distance"
            # encoding, where we will now use two extra bytes to encode the match
            # distance, which allows distances up to 65535 to be represented. The full
            # structure of the opcode is LLMMM111 DDDDDDDD DDDDDDDD LITERAL.
            opc_len = 3
            L = _extract(opc, 8, 6, 2)
            M = _extract(opc, 8, 3, 3) + 3

            if src_size <= opc_len + L:
                break

            (D,) = _H.unpack(src.read(2))

        elif opc in OP_PRE_D:
            # "previous distance": This opcode has the structure LLMMM110, where the
            # length of the literal (0-3 bytes) is encoded by the high 2 bits of the
            # first byte. We first extract the literal length so we know how long
            # the opcode is, then check that the source can hold both this opcode and
            # at least one byte of the next (because any valid input stream must be
            # terminated with an eos token).
            opc_len = 1
            L = _extract(opc, 8, 6, 2)
            M = _extract(opc, 8, 3, 3) + 3

            if src_size <= opc_len + L:
                break

        elif opc in OP_SML_M:
            # "small match": This opcode has no literal, and uses the previous match
            # distance (i.e. it encodes only the match length), in a single byte as
            # 1111MMMM.
            opc_len = 1
            L = None
            M = _extract(opc, 8, 0, 4)

            if src_size <= opc_len:
                break

        elif opc in OP_LRG_M:
            # "large match": This opcode has no literal, and uses the previous match
            # distance (i.e. it encodes only the match length). It is encoded in two
            # bytes as 11110000 MMMMMMMM.  Because matches smaller than 16 bytes can
            # be represented by sml_m, there is an implicit bias of 16 on the match
            # length; the representable values are [16,271].
            opc_len = 2
            L = None

            if src_size <= opc_len:
                break

            M = src.read(1)[0] + 16

        elif opc in OP_SML_L:
            # "small literal": This opcode has no match, and encodes only a literal
            # of length up to 15 bytes. The format is 1110LLLL LITERAL.
            opc_len = 1
            L = _extract(opc, 8, 0, 4)
            M = None

        elif opc in OP_LRG_L:
            # "large literal": This opcode has no match, and uses the previous match
            # distance (i.e. it encodes only the match length). It is encoded in two
            # bytes as 11100000 LLLLLLLL LITERAL.  Because literals smaller than 16
            # bytes can be represented by sml_l, there is an implicit bias of 16 on
            # the literal length; the representable values are [16,271].
            opc_len = 2

            if src_size <= opc_len:
                break

            L = src.read(1)[0] + 16
            M = None

        elif opc in OP_NOP:
            opc_len = 1
            L = None
            M = None

            if src_size <= opc_len:
                break

        elif opc in OP_EOS:
            opc_len = 8

            if src_size < opc_len:
                break

            src_size -= opc_len + L
            break

        elif opc in OP_UDEF:
            raise ValueError("Undefined opcode")

        # Update remaining source size
        src_size -= opc_len

        # Copy literal
        if L is not None:
            src_size -= L
            dst += src.read(L)

        # Match
        if M is not None:
            if len(dst) < D or D == 0:
                raise ValueError("Invalid match distance")

            remaining = M
            while remaining > 0:
                match_size = min(remaining, D)
                dst += dst[-D : (-D + match_size) or None]
                remaining -= match_size

    return bytes(dst)


def _extract(container: int, container_width: int, lsb: int, width: int) -> int:
    if width == container_width:
        return container
    return (container >> lsb) & ((1 << width) - 1)



from __future__ import annotations

from dissect.util.compression import lz4, lzo

lz4_python = lz4
lzo_python = lzo

# This selects between a native Rust version of lz4 and lzo (when available) and our own
# pure-Python implementation.
#
# By doing a:
#  from dissect.util.compression import lz4
# or
#  from dissect.util.compression import lzo
#
# in another project will automatically give you one or the other.
#
# The native Rust version is also available as dissect.util.compression.lz4_native
# and dissect.util.compression.lzo_native (when available) and the pure Python
# version is always available as dissect.util.compression.lz4_python and
# dissect.util.compression.lzo_python.
#
# Note that the pure Python implementation and the Rust implementation are NOT a full replacement
# for the "official" lz4 and lzo Python packages: only the decompress() function is implemented.
try:
    from dissect.util import _native

    lz4 = lz4_native = _native.compression.lz4
    lzo = lzo_native = _native.compression.lzo
except (ImportError, AttributeError):
    lz4_native = lzo_native = None

__all__ = [
    "lz4",
    "lz4_native",
    "lz4_python",
    "lzbitmap",
    "lzfse",
    "lznt1",
    "lzo",
    "lzo_native",
    "lzo_python",
    "lzvn",
    "lzxpress",
    "lzxpress_huffman",
    "sevenbit",
]



from __future__ import annotations

import struct
import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterator


def xmemoryview(view: bytes, format: str) -> memoryview | _xmemoryview:
    """Cast a memoryview to the specified format, including endianness.

    The regular ``memoryview.cast()`` method only supports host endianness. While that should be fine 99% of the time
    (most of the world runs on little endian systems), we'd rather it be fine 100% of the time. This utility method
    ensures that by transparently converting between endianness if it doesn't match the host endianness.

    If the host endianness matches the requested endianness, this simply returns a regular ``memoryview.cast()``.

    See ``memoryview.cast()`` for more details on what that actually does.

    Args:
        view: The bytes object or memoryview to cast.
        format: The format to cast to in ``struct`` format syntax.

    Raises:
        ValueError: If the format is invalid.
        TypeError: If the view is of an invalid type.
    """
    if len(format) != 2:
        raise ValueError("Invalid format specification")

    if isinstance(view, bytes | bytearray):
        view = memoryview(view)

    if not isinstance(view, memoryview):
        raise TypeError("view must be a memoryview, bytes or bytearray object")

    endian = format[0]
    view = view.cast(format[1])

    if (
        endian in ("@", "=")
        or (sys.byteorder == "little" and endian == "<")
        or (sys.byteorder == "big" and endian in (">", "!"))
    ):
        # Native endianness, don't need to do anything
        return view

    # Non-native endianness
    return _xmemoryview(view, format)


class _xmemoryview:
    """Wrapper for memoryview that converts between host and a different destination endianness.

    Args:
        view: The (already casted) memoryview to wrap.
        format: The format to convert to.
    """

    def __init__(self, view: memoryview, format: str):
        self._format = format

        fmt = format[1]
        self._view = view
        self._struct_frm = struct.Struct(f"={fmt}")
        self._struct_to = struct.Struct(format)

    def tolist(self) -> list[int]:
        return self._convert_from_native(self._view.tolist())

    def _convert_from_native(self, value: list[int] | int) -> int:
        if isinstance(value, list):
            endian = self._format[0]
            fmt = self._format[1]
            pck = f"{len(value)}{fmt}"
            return list(struct.unpack(f"{endian}{pck}", struct.pack(f"={pck}", *value)))
        return self._struct_to.unpack(self._struct_frm.pack(value))[0]

    def _convert_to_native(self, value: list[int] | int) -> int:
        if isinstance(value, list):
            endian = self._format[0]
            fmt = self._format[1]
            pck = f"{len(value)}{fmt}"
            return list(struct.unpack(f"={pck}", struct.pack(f"{endian}{pck}", *value)))
        return self._struct_frm.unpack(self._struct_to.pack(value))[0]

    def __getitem__(self, idx: int | slice) -> int | bytes:
        value = self._view[idx]
        if isinstance(idx, int):
            return self._convert_from_native(value)
        if isinstance(idx, slice):
            return _xmemoryview(self._view[idx], self._format)

        raise TypeError("Invalid index type")

    def __setitem__(self, idx: int | slice, value: list[int] | int) -> None:
        if isinstance(idx, int | slice):
            self._view[idx] = self._convert_to_native(value)
        else:
            raise TypeError("Invalid index type")

    def __len__(self) -> int:
        return len(self._view)

    def __eq__(self, other: memoryview | _xmemoryview) -> bool:
        if isinstance(other, _xmemoryview):
            other = other._view
        return self._view.__eq__(other)

    def __iter__(self) -> Iterator[int]:
        for value in self._view:
            yield self._convert_from_native(value)

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._view, attr)



from __future__ import annotations

import io
import struct
from typing import BinaryIO


def read_sid(fh: BinaryIO | bytes, endian: str = "<", swap_last: bool = False) -> str:
    """Read a Windows SID from bytes.

    Normally we'd do this with cstruct, but do it with just struct to keep dissect.util dependency-free.
    On the upside, this also improves performance!

    This is equivalent to the following structure::

        typedef struct _SID {
            BYTE        Revision;
            BYTE        SubAuthorityCount;
            CHAR        IdentifierAuthority[6];
            DWORD       SubAuthority[SubAuthorityCount];
        } SID;

    Args:
        fh: A file-like object or bytes object to read the SID from.
        endian: Optional endianness for reading the sub authorities.
        swap_list: Optional flag for swapping the endianess of the _last_ sub authority entry.
    """
    if isinstance(fh, bytes):
        fh = io.BytesIO(fh)

    if len(buf := fh.read(8)) != 8:
        return ""

    revision = buf[0]
    sub_authority_count = buf[1]
    authority = int.from_bytes(buf[2:], "big")

    sub_authority_buf = bytearray(fh.read(sub_authority_count * 4))
    if sub_authority_count and swap_last:
        sub_authority_buf[-4:] = sub_authority_buf[-4:][::-1]

    sub_authorities = struct.unpack(f"{endian}{sub_authority_count}I", sub_authority_buf)

    sid_elements = [
        "S",
        f"{revision}",
        f"{authority}",
    ]
    sid_elements.extend(map(str, sub_authorities))
    return "-".join(sid_elements)


def write_sid(sid: str, endian: str = "<", swap_last: bool = False) -> bytes:
    """Write a Windows SID string to bytes.

    Args:
        sid: SID in the form ``S-Revision-Authority-SubAuth1-...``.
        endian: Optional endianness for reading the sub authorities.
        swap_last: Optional flag for swapping the endianess of the _last_ sub authority entry.
    """
    if not sid:
        return b""

    parts = sid.split("-")
    if len(parts) < 3 or parts[0].upper() != "S":
        raise ValueError("Invalid SID string format: insufficient parts")

    revision = int(parts[1]).to_bytes(1, "little")
    authority = int(parts[2]).to_bytes(6, "big")
    sub_authorities = [int(x) for x in parts[3:]]

    header = revision + len(sub_authorities).to_bytes(1, "little") + authority

    if not sub_authorities:
        return header

    sub_bytes = bytearray(struct.pack(f"{endian}{len(sub_authorities)}I", *sub_authorities))
    if swap_last:
        sub_bytes[-4:] = sub_bytes[-4:][::-1]

    return header + bytes(sub_bytes)



from __future__ import annotations

from dissect.util.hash import crc32c

crc32c_python = crc32c

# This selects between a native Rust version of crc32c (when available) and our own
# pure-Python implementation.
#
# By doing a:
#  from dissect.util.hash import crc32c
#
# in another project will automatically give you one or the other.
#
# The native Rust version is also available as dissect.util.hash.crc32c_native (when available)
# and the pure Python version is always available as dissect.util.hash.crc32c_python.
try:
    from dissect.util import _native

    crc32c = crc32c_native = _native.hash.crc32c
except (ImportError, AttributeError):
    crc32c_native = None

__all__ = [
    "crc32",
    "crc32c",
    "crc32c_native",
    "crc32c_python",
    "jenkins",
]



# Reference: https://github.com/google/rekall/blob/master/rekall-core/rekall/plugins/filesystems/lznt1.py
from __future__ import annotations

import array
import io
import struct
from typing import BinaryIO


def _get_displacement(offset: int) -> int:
    """Calculate the displacement."""
    result = 0
    while offset >= 0x10:
        offset >>= 1
        result += 1

    return result


DISPLACEMENT_TABLE = array.array("B", [_get_displacement(x) for x in range(8192)])

COMPRESSED_MASK = 1 << 15
SIGNATURE_MASK = 3 << 12
SIZE_MASK = (1 << 12) - 1
TAG_MASKS = [(1 << i) for i in range(8)]


def decompress(src: bytes | BinaryIO) -> bytes:
    """LZNT1 decompress from a file-like object or bytes.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    offset = src.tell()
    src.seek(0, io.SEEK_END)
    size = src.tell() - offset
    src.seek(offset)

    dst = io.BytesIO()

    while src.tell() - offset < size:
        block_offset = src.tell()
        uncompressed_chunk_offset = dst.tell()

        block_header = struct.unpack("<H", src.read(2))[0]
        if block_header & SIGNATURE_MASK != SIGNATURE_MASK:
            break

        hsize = block_header & SIZE_MASK

        block_end = block_offset + hsize + 3

        if block_header & COMPRESSED_MASK:
            while src.tell() < block_end:
                header = ord(src.read(1))
                for mask in TAG_MASKS:
                    if src.tell() >= block_end:
                        break

                    if header & mask:
                        pointer = struct.unpack("<H", src.read(2))[0]
                        displacement = DISPLACEMENT_TABLE[dst.tell() - uncompressed_chunk_offset - 1]

                        symbol_offset = (pointer >> (12 - displacement)) + 1
                        symbol_length = (pointer & (0xFFF >> displacement)) + 3

                        dst.seek(-symbol_offset, io.SEEK_END)
                        data = dst.read(symbol_length)

                        # Pad the data to make it fit.
                        if 0 < len(data) < symbol_length:
                            data = data * (symbol_length // len(data) + 1)
                            data = data[:symbol_length]

                        dst.seek(0, io.SEEK_END)
                        dst.write(data)
                    else:
                        data = src.read(1)
                        dst.write(data)

        else:
            # Block is not compressed
            data = src.read(hsize + 1)
            dst.write(data)

    return dst.getvalue()



from __future__ import annotations

import operator
import re
from enum import Enum

from dissect.util.exceptions import Error


class InvalidQueryError(Error):
    pass


class LogicalOperator(Enum):
    AND = "&"
    OR = "|"
    NOT = "!"


_LOGICAL_OPERATORS = tuple(op.value for op in LogicalOperator)


class ComparisonOperator(Enum):
    GE = ">="
    LE = "<="
    GT = ">"
    LT = "<"
    EQ = "="
    APPROX = "~="
    BIT = ":="
    EXTENDED = ":"


_NORMAL_COMPARISON_OPERATORS = [op for op in ComparisonOperator if op != ComparisonOperator.EXTENDED]
_SORTED_COMPARISON_OPERATORS = sorted(_NORMAL_COMPARISON_OPERATORS, key=lambda op: len(op.value), reverse=True)

_RE_EXTENDED = re.compile(r"(.+?):(.+?):=(.+)?")


class SearchFilter:
    """Represents an LDAP search filter (simple or nested).

    Args:
        query: The LDAP search filter string.
    """

    def __init__(self, query: str) -> None:
        self.query: str = query

        self.children: list[SearchFilter] = []
        self.operator: LogicalOperator | ComparisonOperator | None = None
        self.attribute: str | None = None
        self.value: str | None = None
        self._extended_rule: str | None = None

        _validate_syntax(query)

        if query[1:-1].startswith(_LOGICAL_OPERATORS):
            self._parse_nested()
        else:
            self._parse_simple()

    def __repr__(self) -> str:
        if self.is_nested():
            return f"<SearchFilter nested operator={self.operator.value!r} children={self.children}>"
        return f"<SearchFilter attribute={self.attribute!r} operator={self.operator.value!r} value={self.value}>"

    @classmethod
    def parse(cls, query: str, optimize: bool = True) -> SearchFilter:
        """Parse an LDAP query into a filter object, with optional optimization."""
        result = cls(query)
        if optimize:
            return optimize_ldap_query(result)[0]
        return result

    def is_nested(self) -> bool:
        """Return whether the filter is nested (i.e., contains logical operators and child filters)."""
        return isinstance(self.operator, LogicalOperator)

    def format(self) -> str:
        """Format the search filter back into an LDAP query string."""
        if self.is_nested():
            childs = "".join([child.format() for child in self.children])
            return f"({self.operator.value}{childs})"

        if self.operator == ComparisonOperator.EXTENDED:
            return f"({self.attribute}:{self._extended_rule}:={self.value})"

        return f"({self.attribute}{self.operator.value}{self.value})"

    def _parse_simple(self) -> None:
        """Parse simple filter."""
        query = self.query[1:-1]

        # Check for extended matching rules first
        if ":" in query and (match := _RE_EXTENDED.match(query)):
            self.operator = ComparisonOperator.EXTENDED
            self.attribute, self._extended_rule, self.value = match.groups()
            return

        # Regular operator parsing
        test = query
        operators: list[ComparisonOperator] = []
        for op in _SORTED_COMPARISON_OPERATORS:
            if op.value not in test:
                continue

            if test.count(op.value) > 1:
                raise InvalidQueryError(f"Comparison operator {op.value} found multiple times in query: {self.query}")

            operators.append(op)
            test = test.replace(op.value, "")

        if len(operators) == 0:
            raise InvalidQueryError(
                f"No comparison operator found in query: {self.query}. "
                f"Expected one of {[op.value for op in _NORMAL_COMPARISON_OPERATORS]}."
            )

        if len(operators) > 1:
            raise InvalidQueryError(
                f"Multiple comparison operators found in query: {self.query} -> {[o.value for o in operators]} "
                f"Expected only one of {[op.value for op in _NORMAL_COMPARISON_OPERATORS]}."
            )

        self.operator = operators[0]
        self.attribute, _, self.value = query.partition(self.operator.value)

    def _parse_nested(self) -> None:
        """Parse nested filter."""
        query = self.query[1:-1]
        self.operator = LogicalOperator(query[0])

        start = 1
        while start < len(query):
            end = start + 1
            depth = 1

            while end < len(query) and depth > 0:
                if query[end] == "(":
                    depth += 1
                elif query[end] == ")":
                    depth -= 1
                end += 1

            self.children.append(SearchFilter(query[start:end]))
            start = end


_ATTRIBUTE_WEIGHTS = {
    "objectGUID": 1,
    "distinguishedName": 1,
    "sAMAccountName": 2,
    "userPrincipalName": 2,
    "mail": 2,
    "sAMAccountType": 3,
    "servicePrincipalName": 3,
    "userAccountControl": 4,
    "memberOf": 5,
    "member": 5,
    "pwdLastSet": 5,
    "primaryGroupID": 6,
    "whenCreated": 6,
    "ou": 6,
    "lastLogonTimestamp": 6,
    "cn": 7,
    "givenName": 7,
    "name": 7,
    "telephoneNumber": 7,
    "objectCategory": 8,
    "description": 9,
    "objectClass": 10,
}


def optimize_ldap_query(query: SearchFilter) -> tuple[SearchFilter, int]:
    """Optimize an LDAP query in-place.

    Removes redundant conditions and sorts filters and conditions based on how specific they are.

    Args:
        query: The LDAP query to optimize.

    Returns:
        A tuple containing the optimized LDAP query and its weight.
    """
    # Simplify single-child AND/OR
    if query.is_nested() and len(query.children) == 1 and query.operator in (LogicalOperator.AND, LogicalOperator.OR):
        return optimize_ldap_query(query.children[0])

    # Sort nested children by weight
    if query.is_nested() and len(query.children) > 1:
        children = sorted((optimize_ldap_query(child) for child in query.children), key=operator.itemgetter(1))

        query.children = [child for child, _ in children]
        query.query = query.format()

        return query, max(weight for _, weight in children)

    # Handle NOT
    if query.is_nested() and len(query.children) == 1 and query.operator == LogicalOperator.NOT:
        child, weight = optimize_ldap_query(query.children[0])

        query.children[0] = child
        query.query = query.format()

        return query, weight

    # Base case: simple filter
    if not query.is_nested():
        return query, _ATTRIBUTE_WEIGHTS.get(query.attribute, max(_ATTRIBUTE_WEIGHTS.values()))

    return query, max(_ATTRIBUTE_WEIGHTS.values())


def _validate_syntax(query: str) -> None:
    """Validate basic LDAP query syntax.

    Args:
        query: The LDAP query to validate.
    """
    if not query:
        raise InvalidQueryError("Empty query")

    if not query.startswith("(") or not query.endswith(")"):
        raise InvalidQueryError(f"Query must be wrapped in parentheses: {query}")

    if query.count("(") != query.count(")"):
        raise InvalidQueryError(f"Unbalanced parentheses in query: {query}")

    # Check for empty parentheses
    if "()" in query:
        raise InvalidQueryError(f"Empty parentheses found in query: {query}")

    # Check for queries that start with double opening parentheses
    if query.startswith("(("):
        raise InvalidQueryError(f"Invalid query structure: {query}")



from __future__ import annotations

import codecs


def error_handler(error: Exception) -> tuple[str, int]:
    if not isinstance(error, UnicodeDecodeError):
        raise error

    result = []
    for i in range(error.start, error.end):
        byte = error.object[i]
        if byte < 128:
            raise error
        result.append(chr(0xDC00 + byte))

    return "".join(result), error.end


try:
    codecs.lookup_error("surrogateescape")
except LookupError:
    codecs.register_error("surrogateescape", error_handler)



from __future__ import annotations

import io
import struct
from typing import BinaryIO

from dissect.util.exceptions import CorruptDataError


def _get_length(src: BinaryIO, length: int) -> int:
    if length != 0xF:
        return length

    while True:
        read_buf = src.read(1)
        if len(read_buf) != 1:
            raise CorruptDataError("EOF at length read")
        len_part = ord(read_buf)
        length += len_part

        if len_part != 0xFF:
            break

    return length


def decompress(
    src: bytes | BinaryIO,
    uncompressed_size: int = -1,
    return_bytearray: bool = False,
) -> bytes | tuple[bytes, int]:
    """LZ4 decompress from a file-like object or bytes up to a certain length. Assumes no header.

    Args:
        src: File-like object or bytes to decompress from.
        uncompressed_size: Ignored, present for compatibility with native lz4.
        return_bytearray: Whether to return ``bytearray`` or ``bytes``.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    dst = bytearray()
    min_match_len = 4

    while True:
        if len(read_buf := src.read(1)) == 0:
            raise CorruptDataError("EOF at reading literal-len")

        token = ord(read_buf)
        literal_len = _get_length(src, (token >> 4) & 0xF)

        if len(dst) + literal_len > uncompressed_size > 0:
            raise CorruptDataError("Decompressed size exceeds uncompressed_size")

        if len(read_buf := src.read(literal_len)) != literal_len:
            raise CorruptDataError("Not literal data")

        dst.extend(read_buf)
        if len(dst) >= uncompressed_size > 0:
            break

        if len(read_buf := src.read(2)) == 0:
            token_and = token & 0xF
            if token_and != 0:
                raise CorruptDataError(f"EOF, but match-len > 0: {token_and}")
            break

        if len(read_buf) != 2:
            raise CorruptDataError("Premature EOF")

        if (offset := struct.unpack("<H", read_buf)[0]) == 0:
            raise CorruptDataError("Offset can't be 0")

        match_len = _get_length(src, (token >> 0) & 0xF)
        match_len += min_match_len

        if len(dst) + match_len > uncompressed_size > 0:
            raise CorruptDataError("Decompressed size exceeds uncompressed_size")

        remaining = match_len
        while remaining > 0:
            match_size = min(remaining, offset)
            dst += dst[-offset : (-offset + match_size) or None]
            remaining -= match_size

        if len(dst) >= uncompressed_size > 0:
            break

    if not return_bytearray:
        dst = bytes(dst)

    return dst[:uncompressed_size] if uncompressed_size > 0 else dst



# References:
# - https://github.com/google/snappy/blob/main/format_description.txt
from __future__ import annotations

import io
import struct
from typing import BinaryIO

_H = struct.Struct("<H")
_I = struct.Struct("<I")


def varint(src: BinaryIO) -> int:
    result = 0
    shift = 0

    while byte := src.read(1):
        value = byte[0]
        if value < 0x80:
            return result | (value << shift)
        result |= (value & 0x7F) << shift
        shift += 7

    raise EOFError("Unexpected EOF while reading varint")


def decompress(src: bytes | BinaryIO) -> bytes:
    """Snappy decompress from a file-like object or bytes.

    Decompresses until the stored uncompressed length in the preamble.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    dst = io.BytesIO()

    uncompressed_length = varint(src)

    while dst.tell() < uncompressed_length:
        tag_byte = src.read(1)[0]

        if (tag := tag_byte & 0b11) == 0:
            # Literal
            length = tag_byte >> 2
            if length < 60:
                length += 1
            elif length == 60:
                length = src.read(1)[0] + 1
            elif length == 61:
                length = _H.unpack(src.read(2))[0] + 1
            elif length == 62:
                length = _I.unpack(src.read(3) + b"\x00")[0] + 1
            elif length == 63:
                length = _I.unpack(src.read(4))[0] + 1

            if len(buf := src.read(length)) < length:
                raise EOFError("Unexpected EOF while reading literal")

            dst.write(buf)
            continue

        # Copy with 1, 2 or 4 byte offset
        if tag == 1:
            length = ((tag_byte >> 2) & 0b111) + 4
            offset = ((tag_byte & 0b11100000) << 3) | src.read(1)[0]
        elif tag == 2:
            length = (tag_byte >> 2) + 1
            offset = _H.unpack(src.read(2))[0]
        else:
            length = (tag_byte >> 2) + 1
            offset = _I.unpack(src.read(4))[0]

        dst_offset = dst.tell() - offset
        buf = dst.getbuffer()[dst_offset : dst_offset + length].tobytes()
        if offset - length <= 0:
            buf = (buf * ((length // len(buf)) + 1))[:length]

        dst.write(buf)

    return dst.getvalue()



# References:
# - https://github.com/FFmpeg/FFmpeg/blob/master/libavutil/lzo.c
# - https://docs.kernel.org/staging/lzo.html
# - https://github.com/torvalds/linux/blob/master/lib/lzo/lzo1x_decompress_safe.c
from __future__ import annotations

import io
import struct
from typing import BinaryIO

MAX_READ_LENGTH = (1 << 32) - 1000


def _read_length(src: BinaryIO, val: int, mask: int) -> int:
    if (length := val & mask) != 0:
        return length

    while (val := src.read(1)[0]) == 0:
        if length >= MAX_READ_LENGTH:
            raise ValueError("Invalid encoded length")
        length += 255

    return length + mask + val


def decompress(src: bytes | BinaryIO, header: bool = True, buflen: int = -1) -> bytes:
    """LZO decompress from a file-like object or bytes. Assumes no header.

    Arguments are largely compatible with python-lzo API.

    Args:
        src: File-like object or bytes to decompress.
        header: Whether the metadata header is included in the input.
        buflen: If ``header`` is ``False``, a buffer length in bytes must be given that will fit the output.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    dst = bytearray()

    if header:
        byte = src.read(1)[0]
        if byte not in [0xF0, 0xF1]:
            raise ValueError("Invalid header value")
        out_len = struct.unpack("<I", src.read(4))[0]
    else:
        out_len = buflen

    val = src.read(1)[0]
    offset = src.tell()
    if src.seek(5) == 5 and val == 17:
        src.seek(offset)
        _ = src.read(1)  # This would be the bitstream version but we don't currently use it
        val = src.read(1)[0]
    else:
        src.seek(offset)

    if val > 17:
        dst += src.read(val - 17)
        val = src.read(1)[0]

        if val < 16:
            raise ValueError("Invalid LZO stream")

    state = 0
    while True:
        if val > 15:
            if val > 63:
                # Copy 3-8 bytes from block within 2kb distance
                length = (val >> 5) - 1
                dist = (src.read(1)[0] << 3) + ((val >> 2) & 7) + 1
            elif val > 31:
                # Copy of small block within 16kb distance
                length = _read_length(src, val, 31)
                val = src.read(1)[0]
                dist = (src.read(1)[0] << 6) + (val >> 2) + 1
            else:
                # Copy of a block within 16...48kB distance
                length = _read_length(src, val, 7)
                dist = (1 << 14) + ((val & 8) << 11)
                val = src.read(1)[0]
                dist += (src.read(1)[0] << 6) + (val >> 2)
                if dist == (1 << 14):
                    if length != 1:
                        raise ValueError("Invalid LZO stream")
                    break
        elif not state:
            # Copy 4 or more literals, depending on the last 4 bits
            length = _read_length(src, val, 15)
            dst += src.read(length + 3)
            val = src.read(1)[0]
            if val > 15:
                continue
            length = 1
            dist = (1 << 11) + (src.read(1)[0] << 2) + (val >> 2) + 1
        else:
            length = 0
            dist = (src.read(1)[0] << 2) + (val >> 2) + 1

        remaining = length + 2
        while remaining > 0:
            match_size = min(remaining, dist)
            dst += dst[-dist : (-dist + match_size) or None]
            remaining -= match_size

        # State is often encoded in the last 2 bits of the value, and used in subsequent iterations
        state = length = val & 3
        dst += src.read(length)

        if len(dst) == out_len:
            break

        val = src.read(1)[0]

    return bytes(dst)



from __future__ import annotations

from dissect.util.compression import (
    lz4,
    lznt1,
    lzo,
    lzxpress,
    lzxpress_huffman,
    sevenbit,
)
from dissect.util.hash import crc32c, jenkins

__all__ = [
    "crc32c",
    "jenkins",
    "lz4",
    "lznt1",
    "lzo",
    "lzxpress",
    "lzxpress_huffman",
    "sevenbit",
]



from __future__ import annotations

import argparse
from typing import Any

from dissect.util.plist import NSKeyedArchiver, NSObject


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=argparse.FileType("rb"), help="NSKeyedArchiver plist file to dump")
    args = parser.parse_args()

    with args.file as fh:
        try:
            obj = NSKeyedArchiver(fh)
        except ValueError as e:
            parser.exit(str(e))

        print(obj)
        print_object(obj.top)


def print_object(obj: Any, indent: int = 0, seen: set | None = None) -> None:
    if seen is None:
        seen = set()

    try:
        if obj in seen:
            print(fmt(f"Recursive -> {obj}", indent))
            return
    except Exception:  # noqa: S110
        pass

    if isinstance(obj, list):
        for i, v in enumerate(obj):
            print(fmt(f"[{i}]:", indent))
            print_object(v, indent + 1, seen)

    elif isinstance(obj, dict | NSObject):
        if isinstance(obj, NSObject):
            print(fmt(obj, indent))
            try:
                seen.add(obj)
            except TypeError:
                pass

        for k in sorted(obj.keys()):
            print(fmt(f"{k}:", indent + 1))
            print_object(obj[k], indent + 2, seen)

    else:
        print(fmt(obj, indent))


def fmt(obj: Any, indent: int) -> str:
    return f"{' ' * (indent * 4)}{obj}"


if __name__ == "__main__":
    main()



# References:
# - https://github.com/lzfse/lzfse
from __future__ import annotations

import io
import struct
from typing import BinaryIO, NamedTuple

from dissect.util.compression import lzvn

LZFSE_ENDOFSTREAM_BLOCK_MAGIC = b"bvx$"  # 0x24787662 (end of stream)
LZFSE_UNCOMPRESSED_BLOCK_MAGIC = b"bvx-"  # 0x2d787662 (raw data)
LZFSE_COMPRESSEDV1_BLOCK_MAGIC = b"bvx1"  # 0x31787662 (lzfse compressed, uncompressed tables)
LZFSE_COMPRESSEDV2_BLOCK_MAGIC = b"bvx2"  # 0x32787662 (lzfse compressed, compressed tables)
LZFSE_COMPRESSEDLZVN_BLOCK_MAGIC = b"bvxn"  # 0x6e787662 (lzvn compressed)

# Throughout LZFSE we refer to "L", "M" and "D"; these will always appear as
# a triplet, and represent a "usual" LZ-style literal and match pair.  "L"
# is the number of literal bytes, "M" is the number of match bytes, and "D"
# is the match "distance"; the distance in bytes between the current pointer
# and the start of the match.
LZFSE_ENCODE_L_SYMBOLS = 20
LZFSE_ENCODE_M_SYMBOLS = 20
LZFSE_ENCODE_D_SYMBOLS = 64
LZFSE_ENCODE_LITERAL_SYMBOLS = 256
LZFSE_ENCODE_L_STATES = 64
LZFSE_ENCODE_M_STATES = 64
LZFSE_ENCODE_D_STATES = 256
LZFSE_ENCODE_LITERAL_STATES = 1024
LZFSE_MATCHES_PER_BLOCK = 10000
LZFSE_LITERALS_PER_BLOCK = 4 * LZFSE_MATCHES_PER_BLOCK

# fmt: off
_lzfse_freq_nbits_table = (
    2, 3, 2, 5, 2, 3, 2, 8, 2, 3, 2, 5, 2, 3, 2, 14,
    2, 3, 2, 5, 2, 3, 2, 8, 2, 3, 2, 5, 2, 3, 2, 14
)
_lzfse_freq_value_table = (
    0, 2, 1, 4, 0, 3, 1, -1, 0, 2, 1, 5, 0, 3, 1, -1,
    0, 2, 1, 6, 0, 3, 1, -1, 0, 2, 1, 7, 0, 3, 1, -1
)

# The L, M, D data streams are all encoded as a "base" value, which is
# FSE-encoded, and an "extra bits" value, which is the difference between
# value and base, and is simply represented as a raw bit value (because it
# is the low-order bits of a larger number, not much entropy can be
# extracted from these bits by more complex encoding schemes). The following
# tables represent the number of low-order bits to encode separately and the
# base values for each of L, M, and D.
_l_extra_bits = (
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 5, 8
)
_l_base_value = (
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 20, 28, 60
)
_m_extra_bits = (
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 5, 8, 11
)
_m_base_value = (
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 24, 56, 312
)
_d_extra_bits = (
    0,  0,  0,  0,  1,  1,  1,  1,  2,  2,  2,  2,  3,  3,  3,  3,
    4,  4,  4,  4,  5,  5,  5,  5,  6,  6,  6,  6,  7,  7,  7,  7,
    8,  8,  8,  8,  9,  9,  9,  9,  10, 10, 10, 10, 11, 11, 11, 11,
    12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15
)
_d_base_value = (
    0,      1,      2,      3,     4,     6,     8,     10,    12,    16,
    20,     24,     28,     36,    44,    52,    60,    76,    92,    108,
    124,    156,    188,    220,   252,   316,   380,   444,   508,   636,
    764,    892,    1020,   1276,  1532,  1788,  2044,  2556,  3068,  3580,
    4092,   5116,   6140,   7164,  8188,  10236, 12284, 14332, 16380, 20476,
    24572,  28668,  32764,  40956, 49148, 57340, 65532, 81916, 98300, 114684,
    131068, 163836, 196604, 229372
)
# fmt: on


_I = struct.Struct("<I")


def _clz(n: int) -> int:
    if n == 0:
        return 32
    return 32 - n.bit_length()


class LZFSECompressedBlockHeader(NamedTuple):
    """LZFSE compressed block header."""

    __struct_v1__ = struct.Struct(
        "<IIIIIIi4HiHHH"
        f"{LZFSE_ENCODE_L_SYMBOLS}H"
        f"{LZFSE_ENCODE_M_SYMBOLS}H"
        f"{LZFSE_ENCODE_D_SYMBOLS}H"
        f"{LZFSE_ENCODE_LITERAL_SYMBOLS}H"
    )
    __struct_v2__ = struct.Struct("<I3Q")

    n_raw_bytes: int
    n_payload_bytes: int
    n_literals: int
    n_matches: int
    n_literal_payload_bytes: int
    n_lmd_payload_bytes: int
    literal_bits: int
    literal_state: tuple[int, int, int, int]
    lmd_bits: int
    l_state: int
    m_state: int
    d_state: int
    l_freq: tuple[int, ...]
    m_freq: tuple[int, ...]
    d_freq: tuple[int, ...]
    literal_freq: tuple[int, ...]

    @classmethod
    def parse_v1(cls, fh: BinaryIO) -> LZFSECompressedBlockHeader:
        """Decode all fields from a v1 header."""
        values = cls.__struct_v1__.unpack(fh.read(cls.__struct_v1__.size))

        return cls(
            n_raw_bytes=values[0],
            n_payload_bytes=values[1],
            n_literals=values[2],
            n_matches=values[3],
            n_literal_payload_bytes=values[4],
            n_lmd_payload_bytes=values[5],
            literal_bits=values[6],
            literal_state=values[7:11],
            lmd_bits=values[11],
            l_state=values[12],
            m_state=values[13],
            d_state=values[14],
            l_freq=values[15 : 15 + LZFSE_ENCODE_L_SYMBOLS],
            m_freq=values[35 : 35 + LZFSE_ENCODE_M_SYMBOLS],
            d_freq=values[55 : 55 + LZFSE_ENCODE_D_SYMBOLS],
            literal_freq=values[119 : 119 + LZFSE_ENCODE_LITERAL_SYMBOLS],
        )

    @classmethod
    def parse_v2(cls, fh: BinaryIO) -> LZFSECompressedBlockHeader:
        """Decode all fields from a v2 header."""
        values = cls.__struct_v2__.unpack(fh.read(cls.__struct_v2__.size))
        v0, v1, v2 = values[1:4]

        n_literal_payload_bytes = _get_field(v0, 20, 20)
        n_lmd_payload_bytes = _get_field(v1, 40, 20)
        n_payload_bytes = n_literal_payload_bytes + n_lmd_payload_bytes

        freq_tables_size = _get_field(v2, 0, 32) - cls.__struct_v2__.size - 4  # exclude magic

        if freq_tables_size == 0:
            l_freq = (0,) * 20
            m_freq = (0,) * 20
            d_freq = (0,) * 64
            literal_freq = (0,) * 256
        else:
            accum = 0
            accum = int.from_bytes(fh.read(freq_tables_size), "little")
            result = [0] * 720

            for i in range(
                LZFSE_ENCODE_L_SYMBOLS + LZFSE_ENCODE_M_SYMBOLS + LZFSE_ENCODE_D_SYMBOLS + LZFSE_ENCODE_LITERAL_SYMBOLS
            ):
                # Decode and store value
                nbits, value = _decode_v1_freq_value(accum)
                result[i] = value

                # Consume nbits bits
                accum >>= nbits

            l_freq = tuple(result[0:20])
            m_freq = tuple(result[20:40])
            d_freq = tuple(result[40:104])
            literal_freq = tuple(result[104:360])

        return cls(
            n_raw_bytes=values[0],
            n_payload_bytes=n_payload_bytes,
            n_literals=_get_field(v0, 0, 20),
            n_matches=_get_field(v0, 40, 20),
            n_literal_payload_bytes=n_literal_payload_bytes,
            n_lmd_payload_bytes=n_lmd_payload_bytes,
            literal_bits=_get_field(v0, 60, 3) - 7,
            literal_state=(
                _get_field(v1, 0, 10),
                _get_field(v1, 10, 10),
                _get_field(v1, 20, 10),
                _get_field(v1, 30, 10),
            ),
            lmd_bits=_get_field(v1, 60, 3) - 7,
            l_state=_get_field(v2, 32, 10),
            m_state=_get_field(v2, 42, 10),
            d_state=_get_field(v2, 52, 10),
            l_freq=l_freq,
            m_freq=m_freq,
            d_freq=d_freq,
            literal_freq=literal_freq,
        )


def _get_field(value: int, offset: int, nbits: int) -> int:
    """Extracts up to 32 bits from a 64-bit field beginning at offset, and zero-extends them to a 32-bit int.

    If we number the bits of ``value`` from 0 (least significant) to 63 (most significant),
    the result is bits ``offset`` to ``offset+nbits-1``.
    """
    return (value >> offset) & ((1 << nbits) - 1)


def _decode_v1_freq_value(bits: int) -> tuple[int, int]:
    """Decode an entry value from next bits of stream."""
    b = bits & 31
    n = _lzfse_freq_nbits_table[b]

    # Special cases for > 5 bits encoding
    if n == 8:
        value = 8 + ((bits >> 4) & 0xF)
    elif n == 14:
        value = 24 + ((bits >> 4) & 0x3FF)
    else:
        value = _lzfse_freq_value_table[b]

    return n, value


class DecoderEntry(NamedTuple):
    """Entry for one state in the decoder table."""

    k: int  # Number of bits to read
    symbol: int  # Emitted symbol
    delta: int  # Signed increment used to compute next state (+bias)


class ValueDecoderEntry(NamedTuple):
    """Entry for one state in the value decoder table."""

    total_bits: int  # state bits + extra value bits = shift for next decode
    value_bits: int  # extra value bits
    delta: int  # state base (delta)
    vbase: int  # value base


def _init_decoder_table(nstates: int, freq: tuple[int, ...]) -> list[DecoderEntry]:
    """Initialize decoder table ``T[nstates]``.

    ``nstates = sum freq[i]`` is the number of states (a power of 2).
    ``freq`` is a normalized histogram of symbol frequencies, with ``freq[i] >= 0``.

    Some symbols may have a 0 frequency. In that case, they should not be present in the data.
    """
    table = []

    n_clz = _clz(nstates)
    sum_of_freq = 0

    for i, f in enumerate(freq):
        if f == 0:
            # skip this symbol, no occurrences
            continue

        sum_of_freq += f
        if sum_of_freq > nstates:
            raise ValueError("Invalid frequency table")
        k = _clz(f) - n_clz  # shift needed to ensure N <= (F<<K) < 2*N
        j0 = ((2 * nstates) >> k) - f

        # Initialize all states S reached by this symbol: OFFSET <= S < OFFSET + F
        table.extend(
            DecoderEntry(
                k=k if j < j0 else k - 1,
                symbol=i,
                delta=((f + j) << k) - nstates if j < j0 else (j - j0) << (k - 1),
            )
            for j in range(f)
        )

    return table


def _init_value_decoder_table(
    nstates: int, freq: tuple[int, ...], symbol_vbits: tuple[int, ...], symbol_vbase: tuple[int, ...]
) -> list[ValueDecoderEntry]:
    """Initialize value decoder table ``T[nstates]``.

    ``nstates = sum freq[i]`` is the number of states (a power of 2).
    ``freq`` is a normalized histogram of symbol frequencies, with ``freq[i] >= 0``.
    ``symbol_vbits`` and ``symbol_vbase`` are the number of value bits to read and the base value for each symbol.

    Some symbols may have a 0 frequency. In that case, they should not be present in the data.
    """
    table = []

    n_clz = _clz(nstates)
    for i, f in enumerate(freq):
        if f == 0:
            # skip this symbol, no occurrences
            continue

        k = _clz(f) - n_clz  # shift needed to ensure N <= (F<<K) < 2*N
        j0 = ((2 * nstates) >> k) - f

        # Initialize all states S reached by this symbol: OFFSET <= S < OFFSET + F
        table.extend(
            ValueDecoderEntry(
                total_bits=k + symbol_vbits[i] if j < j0 else (k - 1) + symbol_vbits[i],
                value_bits=symbol_vbits[i],
                delta=(((f + j) << k) - nstates) if j < j0 else ((j - j0) << (k - 1)),
                vbase=symbol_vbase[i],
            )
            for j in range(f)
        )

    return table


class _BitStream:
    def __init__(self, data: bytes, nbits: int):
        self.accum = int.from_bytes(data, "little")
        self.nbits = nbits + (len(data) * 8)

    def pull(self, n: int) -> int:
        self.nbits -= n
        result = self.accum >> self.nbits
        self.accum &= (1 << self.nbits) - 1
        return result


def _decode(state: int, decoder_table: list[DecoderEntry], in_stream: _BitStream) -> tuple[int, int]:
    """Decode and return symbol using the decoder table, and update state."""
    e = decoder_table[state]

    # Update state from K bits of input + DELTA
    state = e.delta + in_stream.pull(e.k)

    # Return the symbol for this state
    return state, e.symbol


def _value_decode(state: int, decoder_table: list[ValueDecoderEntry], in_stream: _BitStream) -> tuple[int, int]:
    """Decode and return value using the decoder table, and update state."""
    entry = decoder_table[state]

    state_and_value_bits = in_stream.pull(entry.total_bits)
    state = entry.delta + (state_and_value_bits >> entry.value_bits)

    return state, entry.vbase + (state_and_value_bits & ((1 << entry.value_bits) - 1))


def _decode_lmd(
    header: LZFSECompressedBlockHeader,
    literals: list[int],
    l_decoder: list[ValueDecoderEntry],
    m_decoder: list[ValueDecoderEntry],
    d_decoder: list[ValueDecoderEntry],
    in_stream: _BitStream,
) -> bytes:
    symbols = header.n_matches

    l_state = header.l_state
    m_state = header.m_state
    d_state = header.d_state

    lit = io.BytesIO(bytes(literals))

    L = 0
    M = 0
    D = None

    dst = bytearray()

    while symbols:
        # Decode the next L, M, D symbol from the input stream
        l_state, L = _value_decode(l_state, l_decoder, in_stream)
        if (lit.tell() + L) >= LZFSE_LITERALS_PER_BLOCK + 64:
            raise ValueError("Literal overflow")

        m_state, M = _value_decode(m_state, m_decoder, in_stream)
        d_state, new_d = _value_decode(d_state, d_decoder, in_stream)
        D = new_d if new_d != 0 else D

        if D is None or len(dst) + L < D:
            raise ValueError("Invalid match distance")

        dst += lit.read(L)

        remaining = M
        while remaining > 0:
            match_size = min(remaining, D)
            dst += dst[-D : (-D + match_size) or None]
            remaining -= match_size

        symbols -= 1

    return bytes(dst)


def decompress(src: bytes | BinaryIO) -> bytes:
    """LZFSE decompress from a file-like object or bytes.

    Decompresses until EOF or EOS of the input data.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    dst = bytearray()

    while True:
        magic = src.read(4)

        if magic == LZFSE_ENDOFSTREAM_BLOCK_MAGIC:
            break

        if magic == LZFSE_UNCOMPRESSED_BLOCK_MAGIC:
            (n_raw_bytes,) = _I.unpack(src.read(4))
            if n_raw_bytes == 0:
                continue

            dst += src.read(n_raw_bytes)

        elif magic in (LZFSE_COMPRESSEDV1_BLOCK_MAGIC, LZFSE_COMPRESSEDV2_BLOCK_MAGIC):
            if magic == LZFSE_COMPRESSEDV1_BLOCK_MAGIC:
                header = LZFSECompressedBlockHeader.parse_v1(src)
            else:
                header = LZFSECompressedBlockHeader.parse_v2(src)

            literal_decoder = _init_decoder_table(LZFSE_ENCODE_LITERAL_STATES, header.literal_freq)
            l_decoder = _init_value_decoder_table(LZFSE_ENCODE_L_STATES, header.l_freq, _l_extra_bits, _l_base_value)
            m_decoder = _init_value_decoder_table(LZFSE_ENCODE_M_STATES, header.m_freq, _m_extra_bits, _m_base_value)
            d_decoder = _init_value_decoder_table(LZFSE_ENCODE_D_STATES, header.d_freq, _d_extra_bits, _d_base_value)

            in_stream = _BitStream(src.read(header.n_literal_payload_bytes), header.literal_bits)

            literals = []
            state0, state1, state2, state3 = header.literal_state
            for _ in range(0, header.n_literals, 4):
                state0, result = _decode(state0, literal_decoder, in_stream)
                literals.append(result)
                state1, result = _decode(state1, literal_decoder, in_stream)
                literals.append(result)
                state2, result = _decode(state2, literal_decoder, in_stream)
                literals.append(result)
                state3, result = _decode(state3, literal_decoder, in_stream)
                literals.append(result)

            in_stream = _BitStream(src.read(header.n_lmd_payload_bytes), header.lmd_bits)
            dst += _decode_lmd(header, literals, l_decoder, m_decoder, d_decoder, in_stream)

        elif magic == LZFSE_COMPRESSEDLZVN_BLOCK_MAGIC:
            (n_raw_bytes,) = _I.unpack(src.read(4))
            (n_payload_bytes,) = _I.unpack(src.read(4))

            dst += lzvn.decompress(src.read(n_payload_bytes))

        else:
            raise ValueError(f"Invalid LZFSE block magic: {magic!r}")

    return bytes(dst)



# Reference: [MS-XCA]
from __future__ import annotations

import io
import struct
from typing import BinaryIO


def decompress(src: bytes | BinaryIO) -> bytes:
    """LZXPRESS decompress from a file-like object or bytes.

    Args:
        src: File-like object or bytes to decompress.

    Returns:
        The decompressed data.
    """
    if not hasattr(src, "read"):
        src = io.BytesIO(src)

    offset = src.tell()
    src.seek(0, io.SEEK_END)
    size = src.tell() - offset
    src.seek(offset)

    dst = bytearray()

    buffered_flags = 0
    buffered_flags_count = 0
    last_length_half_byte = 0

    while src.tell() - offset < size:
        if buffered_flags_count == 0:
            buffered_flags = struct.unpack("<I", src.read(4))[0]
            buffered_flags_count = 32

        buffered_flags_count -= 1
        if buffered_flags & (1 << buffered_flags_count) == 0:
            dst.append(ord(src.read(1)))
        else:
            if src.tell() - offset == size:
                break

            match = struct.unpack("<H", src.read(2))[0]
            match_offset, match_length = divmod(match, 8)
            match_offset += 1

            if match_length == 7:
                if last_length_half_byte == 0:
                    last_length_half_byte = src.tell()
                    match_length = ord(src.read(1)) % 16
                else:
                    rewind = src.tell()
                    src.seek(last_length_half_byte)
                    match_length = ord(src.read(1)) // 16
                    src.seek(rewind)
                    last_length_half_byte = 0

                if match_length == 15:
                    match_length = ord(src.read(1))
                    if match_length == 255:
                        match_length = struct.unpack("<H", src.read(2))[0]
                        if match_length == 0:
                            match_length = struct.unpack("<I", src.read(4))[0]

                        if match_length < 15 + 7:
                            raise ValueError("wrong match length")

                        match_length -= 15 + 7
                    match_length += 15
                match_length += 7
            match_length += 3

            remaining = match_length
            while remaining > 0:
                match_size = min(remaining, match_offset)
                dst += dst[-match_offset : (-match_offset + match_size) or None]
                remaining -= match_size

    return bytes(dst)



from __future__ import annotations

from struct import unpack


def _mix64(a: int, b: int, c: int) -> int:
    """Mixes three 64-bit values reversibly."""
    # Implement logical right shift by masking first
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 43)
    b = (b - c - a) ^ (a << 9)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 8)
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 38)
    b = (b - c - a) ^ (a << 23)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 5)
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 35)
    b = (b - c - a) ^ (a << 49)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 11)
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 12)
    b = (b - c - a) ^ (a << 18)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 22)

    # Normalize to 64 bits
    return a & 0xFFFFFFFFFFFFFFFF, b & 0xFFFFFFFFFFFFFFFF, c & 0xFFFFFFFFFFFFFFFF


def lookup8(key: bytes, level: int) -> int:
    """Hashes a variable-length key into a 64-bit value.

    This hash function is used in the ESXi kernel.

    References:
        - http://burtleburtle.net/bob/c/lookup8.c
    """
    a: int = level
    b: int = level
    c: int = 0x9E3779B97F4A7C13  # Golden ratio, arbitrary value
    bytes_left: int = len(key)
    i: int = 0

    # Process the key in 24-byte chunks
    while bytes_left >= 24:
        a += int.from_bytes(key[i : i + 8], "little")
        b += int.from_bytes(key[i + 8 : i + 16], "little")
        c += int.from_bytes(key[i + 16 : i + 24], "little")
        a, b, c = _mix64(a, b, c)
        i += 24
        bytes_left -= 24

    # Handle the last 23 bytes
    c = c + len(key)
    if bytes_left > 0:
        for shift, byte in enumerate(key[i:]):
            if shift < 8:
                a += byte << (shift * 8)
            elif shift < 16:
                b += byte << ((shift - 8) * 8)
            else:
                # c takes 23 - 8 - 8 = 7 bytes (length is added to LSB)
                c += byte << ((shift - 15) * 8)

    _, _, c = _mix64(a, b, c)
    return c


def lookup8_quads(key: bytes, level: int) -> int:
    """Hashes a key consisting of ``num`` 64-bit integers into a 64-bit value.

    This hash function is used in the ESXi kernel, but unlike :func:`lookup8`, this variant is not compatible with
    any of the original ``lookup8.c`` implementations. The difference between this variant and :func:`lookup8`
    is that in the final step, the value of ``c`` is incremented by the number of quads, not the number
    of bytes in the key. While ``hash2`` in ``lookup8.c`` is also optimized for 64-bit aligned keys,
    (and uses the number of quads as argument for the key size, not bytes) it uses the length of the key
    in bytes to increment ``c`` in the final step.

    References:
        - http://burtleburtle.net/bob/c/lookup8.c
        - ``HashFunc_HashQuads``
    """
    num = len(key) // 8
    quads = unpack(f"<{num}Q", key)
    remaining = num

    a = level
    b = level
    c = 0x9E3779B97F4A7C13  # Golden ratio, arbitrary value
    while remaining > 2:
        a += quads[num - remaining]
        b += quads[num - remaining + 1]
        c += quads[num - remaining + 2]
        a, b, c = _mix64(a, b, c)
        remaining -= 3

    # This is the main difference from lookup8:
    # c is incremented by the number of quads, not the length of the key.
    c = c + num
    if remaining == 2:
        a += quads[num - remaining]
        b += quads[num - remaining + 1]

    if remaining == 1:
        a += quads[num - remaining]

    _, _, c = _mix64(a, b, c)
    return c
